"""Includes GBMClassifier class."""
from sklearn.ensemble import GradientBoostingClassifier

from organon.ml.modelling.algorithms.core.abstractions.base_sklearn_classifier import BaseSklearnClassifier


class GBMClassifier(BaseSklearnClassifier[GradientBoostingClassifier]):
    """Wrapper for sklearn GradientBoostingClassifier"""

    @property
    def classifier_class(self):
        return GradientBoostingClassifier



"""Includes BaseFeatureSelectionSettings class."""
import pandas as pd


class BaseFeatureSelectionSettings:
    """Base settings for all feature selection services"""
    def __init__(self, data: pd.DataFrame):
        self.data: pd.DataFrame = data



"""
This module is for logging and includes LogHelper class.
"""
import logging
import sys
import threading
from enum import Enum, auto
from typing import List, Dict, Any


class Levels(Enum):
    """Class for constant values of logging levels """
    CRITICAL = auto()
    ERROR = auto()
    WARNING = auto()
    INFO = auto()
    DEBUG = auto()
    NOTSET = auto()


_LEVELS_MAP = {
    Levels.CRITICAL: logging.CRITICAL,
    Levels.ERROR: logging.ERROR,
    Levels.WARNING: logging.WARNING,
    Levels.INFO: logging.INFO,
    Levels.DEBUG: logging.DEBUG,
    Levels.NOTSET: logging.NOTSET
}
_REVERSE_LEVELS_MAP = {value: key for key, value in _LEVELS_MAP.items()}


class LogCallerInfo:
    """Stores FrameInfo information"""

    def __init__(self):
        self.frame = None
        self.package_name = None
        self.module_name = None
        self.class_name = None
        self.func_name = None
        self.file_name = None
        self.line_number = None


class OrganonFormatter(logging.Formatter):
    """
    Custom logging formatter class
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default_format = args[0]
        self.level_formats = {
            Levels.NOTSET: None,
            Levels.INFO: None,
            Levels.DEBUG: None,
            Levels.WARNING: None,
            Levels.ERROR: None,
            Levels.CRITICAL: None
        }

    def format(self, record):
        """Formats a log according to level of message"""
        # pylint: disable=protected-access
        original_fmt = self._style._fmt
        self._style._fmt = self.__get_format(record.levelno)
        result = super().format(record)
        self._style._fmt = original_fmt
        return result

    def __get_format(self, levelno: int):
        """Returns format corresponding to given level"""
        level = None
        if levelno in _REVERSE_LEVELS_MAP:
            level = _REVERSE_LEVELS_MAP[levelno]
        if level in self.level_formats and self.level_formats[level] is not None:
            return self.level_formats[level]
        return self.default_format

    def set_default_format(self, format_str: str):
        """Sets default format"""
        self.default_format = format_str

    def set_level_format(self, level, format_str: str):
        """Sets format string of a level
        :param level: Level of which format string will be set (Levels.DEBUG, Levels.INFO etc)
        :param format_str: Format string to be set to given level
        :type format_str: str
        """
        if level in self.level_formats:
            self.level_formats[level] = format_str


class LogHelper:
    """
    Includes static methods for logging.
    """
    level = Levels.INFO
    default_format = "[%(asctime)s] [%(threadName)s] %(levelname)-8s %(module_name)s:%(line_number)s - %(message)s"
    default_file = "application.log"
    log_formatter = OrganonFormatter(default_format, "%Y-%m-%d %H:%M:%S")
    global_extras: Dict[str, Any] = {}
    handlers: List[logging.Handler] = []

    INITIALIZED = False
    __init_lock = threading.Lock()

    @staticmethod
    def initialize(root_logger_format_str: str = default_format, file_path: str = default_file,
                   log_to_console: bool = True, root_logger_level: Levels = None, log_to_file: bool = False):
        """
        Adds file and console_apps handlers. Sets minimum log level and log format
        :param root_logger_format_str: Format of logs.
        :type root_logger_format_str: str
        :param file_path: Path to log file.
        :type file_path: str
        :param log_to_console: Set to true if logs should be printed to console_apps
        :type log_to_console: bool
        :param log_to_file: Set to true if logs should be printed to file (application.log by default)
        :type log_to_file: bool
        :param root_logger_level: Minimum level of messages to be logged for root logger.
        :type root_logger_level: int
        """
        with LogHelper.__init_lock:
            if LogHelper.INITIALIZED:
                raise ValueError("LogHelper already initialized!")
            if root_logger_level is not None:
                LogHelper.set_logger_level(root_logger_level)
            LogHelper.set_default_formatter_default_format_str(root_logger_format_str)
            if log_to_file:
                LogHelper.add_file_handler(file_path)
            if log_to_console:
                LogHelper.add_console_handler()
            LogHelper.INITIALIZED = True

    @staticmethod
    def add_file_handler(file_path: str, logger_name: str = None):
        """
        Sets file where logs will appear
        :param file_path: path to log file
        :type file_path: str
        """
        file_handler = logging.FileHandler(file_path, encoding="utf-8")
        LogHelper._add_handler_to_logger(file_handler, logger_name)

    @staticmethod
    def add_console_handler(logger_name: str = None):
        """
        Add console_apps handler to loggers
        """
        stream_handler = logging.StreamHandler()
        LogHelper._add_handler_to_logger(stream_handler, logger_name)

    @staticmethod
    def set_default_formatter_default_format_str(format_str: str):
        """Sets default format of default log formatter"""
        LogHelper.log_formatter.set_default_format(format_str)

    @staticmethod
    def set_default_formatter_format_str_for_level(level, format_str: str):
        """Sets format string of a level
        :param level: Level of which format string will be set (Levels.DEBUG, Levels.INFO etc)
        :param format_str: Format string to be set to given level
        :type format_str: str
        """
        LogHelper.log_formatter.set_level_format(level, format_str)

    @staticmethod
    def add_global_extra(key: str, value):
        """
        Adds a global extra, that is a global value to be logged with every log.
        """
        LogHelper.global_extras[key] = value

    @staticmethod
    def _add_handler_to_logger(handler: logging.Handler, logger_name: str = None, use_default_format: bool = True):
        """Adds extra handlers to loggers"""
        if use_default_format:
            handler.setFormatter(LogHelper.log_formatter)
        logger = LogHelper._get_logger(logger_name)
        logger.addHandler(handler)

    @staticmethod
    def _get_all_loggers():
        """Returns all loggers"""
        return [logging.getLogger(name) for name in logging.root.manager.loggerDict]  # pylint: disable=no-member

    @staticmethod
    def set_logger_level(level: Levels, logger_name: str = None):
        """Sets logger level"""
        logger = LogHelper._get_logger(logger_name)
        logging_module_level = LogHelper._get_level_in_logging_module(level)
        logger.setLevel(logging_module_level)

    @staticmethod
    def _get_root_logger():
        return logging.getLogger(None)

    @staticmethod
    def _get_level_in_logging_module(level: Levels):
        if level in _LEVELS_MAP:
            return _LEVELS_MAP[level]
        raise ValueError("Given level cannot be converted to a logging module level")

    @staticmethod
    def info(message: str):
        """
        Logs a message with level "INFO".
        :param message: Message to be logged.
        :type message: str
        """
        log_caller = get_log_caller_info()
        logger = LogHelper._get_logger(log_caller.module_name)
        logger.info(message, extra=LogHelper._get_extras(log_caller))

    @staticmethod
    def debug(message):
        """
        Logs a message with level "DEBUG".
        :param message: Message to be logged.
        :type message: str
        """
        log_caller = get_log_caller_info()
        logger = LogHelper._get_logger(log_caller.module_name)
        logger.debug(message, extra=LogHelper._get_extras(log_caller))

    @staticmethod
    def warning(message):
        """
        Logs a message with level 'WARNING'.
        :param message: Message to be logged.
        :type message: str
        """
        log_caller = get_log_caller_info()
        logger = LogHelper._get_logger(log_caller.module_name)
        logger.warning(message, extra=LogHelper._get_extras(log_caller))

    @staticmethod
    def error(message):
        """
        Logs a message with level 'ERROR'.
        :param message: Message to be logged.
        :type message: str
        """
        log_caller = get_log_caller_info()
        logger = LogHelper._get_logger(log_caller.module_name)
        logger.error(message, extra=LogHelper._get_extras(log_caller))

    @staticmethod
    def exception(message):
        """
        Logs a message with level 'ERROR' and also logs stack trace. Should only be called in an exception handler.
        :param message: Message to be logged.
        :type message: str
        """
        log_caller = get_log_caller_info()
        logger = LogHelper._get_logger(log_caller.module_name)
        logger.exception(message, extra=LogHelper._get_extras(log_caller))
        # message = ExceptionFormattingHelper.format_exception(exception)

    @staticmethod
    def critical(message):
        """
        Logs a message with level 'CRITICAL'.
        :param message: Message to be logged.
        :type message: str
        """
        log_caller = get_log_caller_info()
        logger = LogHelper._get_logger(log_caller.module_name)
        logger.critical(message, extra=LogHelper._get_extras(log_caller))

    @staticmethod
    def _get_extras(log_caller_info: LogCallerInfo, except_list: List[str] = None):
        """
        Adds extra values while logging to be used in format of log formatter
        :param log_caller_info:
        :param except_list: names of attributes in log_caller_info which should not exist in log message
        :return: dictionary of extra values
        """
        extras = {}
        for attr in log_caller_info.__dict__:
            attr_val = getattr(log_caller_info, attr)
            if attr_val is not None:
                if except_list is not None and attr in except_list:
                    extras[attr] = ""
                else:
                    extras[attr] = attr_val
            else:
                extras[attr] = ""
        extras.update(LogHelper.global_extras)
        return extras

    @staticmethod
    def _get_logger(name: str = None):
        """Creates a logger"""
        return logging.getLogger(name)


if hasattr(sys, '_getframe'):
    currentframe = lambda stack_index: sys._getframe(stack_index + 1)  # pylint: disable=protected-access
else:
    def currentframe(stack_index):
        """Return the frame object for the caller's stack frame."""
        try:
            raise Exception
        except Exception:  # pylint: disable=broad-except
            return sys.exc_info()[stack_index].tb_frame.f_back


def get_log_caller_info(stack_index: int = 2) -> LogCallerInfo:
    """
    Creates a LogCallerInfo instance using stack frame info
    :param stack_index: index of FrameInfo in stack
    :return: Stack Frame info where log is called
    """
    info = LogCallerInfo()
    frm = currentframe(stack_index)

    f_locals = frm.f_locals
    if "self" in f_locals:
        info.class_name = f_locals["self"].__class__.__qualname__
    info.frame = frm
    info.package_name = frm.f_globals["__package__"]
    info.module_name = frm.f_globals["__name__"]
    f_code = frm.f_code
    info.func_name = f_code.co_name
    info.line_number = frm.f_lineno
    info.file_name = f_code.co_filename
    return info



"""Includes FineTuningSettings class."""

from organon.ml.common.enums.optimizer_type import OptimizerType


class FineTuningSettings:
    """Settings for fine tuning in object classification"""

    def __init__(self, epoch: int = 10, early_stopping_patience: int = 2, early_stopping_min_delta: float = 0.001,
                 learning_rate: float = 1e-5, optimizer: OptimizerType = OptimizerType.ADAM):
        self.epoch = epoch
        self.early_stopping_patience = early_stopping_patience
        self.early_stopping_min_delta = early_stopping_min_delta
        self.learning_rate = learning_rate
        self.optimizer = optimizer



"""Includes DataFrame class."""
from typing import List, Any, Union

import pandas as pd

from organon.fl.core.businessobjects.idataframe import IDataFrame
from organon.fl.core.helpers import string_helper, list_helper


class DataFrame(IDataFrame):
    """Pandas DataFrame based custom DataFrame implementing IDataFrame abstract class."""

    def __init__(self, row_count: int = None):
        super().__init__(row_count=row_count)
        self.data_frame = pd.DataFrame()
        self.__columns_marked_for_gc = []

    def set_value(self, value, column_name: str, row_index: int):
        self.data_frame.at[row_index, column_name] = value

    def contains_column(self, column_name: str) -> bool:
        return column_name in self.data_frame.columns

    def try_add(self, column_name: str, data: List[Any], mark_for_garbage_collection: bool = False) -> bool:
        cond1 = string_helper.is_null_or_empty(column_name) or column_name in self.data_frame
        if cond1:
            raise ValueError("Column name is inadmissible")
        cond2 = list_helper.is_null_or_empty(data) or len(data) != self.row_count
        if cond2:
            raise ValueError("Data length should be equal to frame-row-count")
        self.data_frame[column_name] = data
        if mark_for_garbage_collection:
            self.__columns_marked_for_gc.append(column_name)
        return True

    def remove(self, column_names: Union[List[str], str]):
        columns_to_remove = []
        if isinstance(column_names, list):
            columns_to_remove.extend(column_names)
        else:
            columns_to_remove.append(column_names)
        self.data_frame.drop(columns=columns_to_remove, inplace=True)

    def get_value(self, column_name: str, row_index: int = None):
        column = self.data_frame[column_name]
        if row_index is not None:
            return column[row_index]
        return column

    def collect_garbage_columns(self):
        self.remove(self.__columns_marked_for_gc)
        self.__columns_marked_for_gc.clear()

    def get_column_names(self) -> List[str]:
        return list(self.data_frame.columns)

    def get_subset_as_pandas_df(self, indices: List[Any] = None, columns: List[str] = None):
        if indices is None and columns is None:
            raise ValueError("Both indices and columns are None")
        if indices is None:
            return self.data_frame[columns]
        if columns is None:
            return self.data_frame.loc[indices]
        return self.data_frame.loc[indices, columns]

    def copy_as_pandas_df(self):
        return self.data_frame.copy(deep=True)



"""Includes UnivariatePerformanceFeatureReductionSettings class."""
from typing import Optional, List
from pandas import DataFrame
from organon.ml.common.enums.target_type import TargetType
from organon.ml.feature_reduction.settings.objects.base_feature_reduction_settings import BaseFeatureReductionSettings

class UnivariatePerformanceFeatureReductionSettings(BaseFeatureReductionSettings):
    """Settings for univariate performance feature reduction"""

    def __init__(self, data: DataFrame, target_type: TargetType = None,
                 target_column_name: str = None,
                 performance_metric: str = None,
                 univariate_performance_threshold: float = None,
                 random_state: int = None, excluded_columns: Optional[List[str]] = None):
        # pylint: disable=too-many-arguments
        super().__init__(data, excluded_columns)
        self.target_type: TargetType = target_type
        self.target_column_name: str = target_column_name
        self.performance_metric: str = performance_metric
        self.univariate_performance_threshold: float = univariate_performance_threshold
        self.random_state: int = random_state



"""
This module keeps the helper functions for datetime type.
"""
from datetime import datetime, timedelta
from typing import Union

from dateutil.parser import parse
from dateutil.relativedelta import relativedelta

from organon.fl.core.helpers.string_helper import is_null_or_empty


def now() -> datetime:
    """
    Returns the current(now) time as a datetime.
    :return: current time
    """
    return datetime.now()


def format_date(date: datetime, format_str: str = None) -> str:
    """
    Formats and returns the date as a string.
    :param date: datetime object to be formatted
    :param format_str: datetime format string
    :return: string representation of the date
    """
    if format_str is None:
        format_str = "%Y-%m-%d %H:%M:%S"
        return date.strftime(format_str)
    return date.strftime(format_str)


def get_date_from_string(date_string: str, date_format: str = "", iso_format: bool = False) -> datetime:
    """
    Converts a date written as string to datetime
    :param date_string: A date written as string
    :type date_string: str
    :param date_format: Format of the date_string
    :type date_format: str
    :param iso_format: specifies is given date string is in iso format
    :type iso_format bool
    :return: datetime instance with given date
    """
    if is_null_or_empty(date_format):
        if iso_format:
            return datetime.fromisoformat(date_string)
        return parse(date_string)
    return datetime.strptime(date_string, date_format)


# pylint: disable=too-many-arguments
def add_to_date(date: datetime, years: int = 0, months: int = 0, days: int = 0, hours: int = 0,
                minutes: int = 0, seconds: int = 0, milliseconds: int = 0, microseconds: int = 0,
                delta: Union[timedelta, relativedelta] = None) -> datetime:
    """
    Adds given values (1 year, 2 seconds etc., -2 months) to given date
    :param microseconds:
    :param milliseconds:
    :param seconds:
    :param hours:
    :param days:
    :param months:
    :param years:
    :param minutes:
    :param date: Initial date
    :param delta: timedelta or relativedelta to add to date
    :return: Date after addition
    """
    if delta is None:
        delta = get_time_delta(years, months, days, hours, minutes, seconds, milliseconds, microseconds)
    return date + delta


def get_time_delta(years: int = 0, months: int = 0, days: int = 0, hours: int = 0,
                   minutes: int = 0, seconds: int = 0, milliseconds: int = 0,
                   microseconds: int = 0) -> relativedelta or timedelta:
    """Returns a timedelta/relativedelta object according to given date delta values.
    The returned obejct can be used to add a timedelta to a datetime object."""
    if milliseconds != 0:
        microseconds += milliseconds * 1000
    if years == 0 and months == 0:
        delta = timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds,
                          microseconds=microseconds)
    else:
        delta = relativedelta(years=years, months=months, days=days, hours=hours, minutes=minutes, seconds=seconds,
                              microseconds=microseconds)
    return delta


def date_to_milliseconds(date: datetime) -> int:
    """
    Converts a datetime instance to milliseconds
    :param date: date to be converted
    :type date: datetime
    :return: milliseconds
    """
    epoch = datetime.utcfromtimestamp(0)
    return int((date - epoch).total_seconds() * 1000)


def milliseconds_to_date(milliseconds: int) -> datetime:
    """
    Converts milliseconds to datetime
    :param int milliseconds: milliseconds since 01-01-1970 (utc)
    :return: milliseconds
    """
    microseconds = int(milliseconds) * 1000  # convert to int to support numpy.int as argument
    epoch = datetime.utcfromtimestamp(0)
    date = epoch + timedelta(microseconds=microseconds)
    return date


class DateDifference:
    """Stores date difference info"""

    def __init__(self, seconds):
        self.total_seconds = seconds
        self.total_minutes = self.total_seconds / 60
        self.total_hours = self.total_minutes / 60
        self.total_days = self.total_hours / 24


def get_date_difference(date1: datetime, date2: datetime):
    """Returns date difference"""
    return DateDifference((date1 - date2).total_seconds())


def get_date_as_integer(date: datetime) -> int:
    """Returns number of milliseconds since 01.01.1970 to given date"""
    try:
        return date_to_milliseconds(date)
    except Exception as exc:  # pylint: disable=broad-except
        raise ValueError(f"Date {format_date(date)} cannot be converted to integer") from exc


def get_integer_as_date(num_milliseconds: int) -> datetime:
    """Converts given number of milliseconds to datetime
    :param int num_milliseconds: number of milliseconds since 01.01.1970 to given date
    """
    try:
        return milliseconds_to_date(num_milliseconds)
    except Exception as exc:  # pylint: disable=broad-except
        raise ValueError(f"Integer {num_milliseconds} cannot be converted to datetime") from exc



"""Includes VotingEnsembleRegressor class."""
from typing import Dict, Any, List, Union

import numpy as np
import pandas as pd

from organon.ml.modelling.algorithms.core.abstractions.base_regressor import BaseRegressor
from organon.ml.modelling.algorithms.helpers.ensembling_helper import get_voting_regressor_prediction


class VotingEnsembleRegressor(BaseRegressor):
    """VotingEnsembleRegressor"""

    # pylint: disable=no-member
    def _fit(self, train_data: pd.DataFrame, target_data: Union[pd.DataFrame, pd.Series], **kwargs):
        estimators: List[BaseRegressor] = self.estimators
        if estimators is None or len(estimators) == 0:
            raise ValueError("Estimators not given")
        for estimator in estimators:
            if not isinstance(estimator, BaseRegressor):
                raise ValueError("All estimators should be an instance of BaseRegressor")
        if not self.prefit:
            for estimator in estimators:
                estimator.fit(train_data, target_data)

    def _predict(self, data: pd.DataFrame) -> pd.DataFrame:
        estimators: List[BaseRegressor] = self.estimators
        prediction_dfs = [estimator.predict(data) for estimator in estimators]
        predictions = np.asarray([pred.iloc[:, 0] for pred in prediction_dfs]).T
        final_pred = get_voting_regressor_prediction(predictions)
        return pd.DataFrame({prediction_dfs[0].columns[0]: final_pred})

    def _get_params_with_defaults(self) -> Dict[str, Any]:
        return {
            "estimators": None,
            "prefit": False,
        }



"""Includes UserCalculationParams class."""
from typing import List

from organon.idq.services.user_settings.base_user_calculation_params import BaseUserCalculationParams
from organon.idq.services.user_settings.user_partition_info import UserPartitionInfo


class UserMultiPartitionCalcParams(BaseUserCalculationParams):
    """Calculation settings user class"""

    def __init__(self, calc_name, input_source_settings, max_nom_card_count, outlier_param, use_pop_nom_stats,
                 partitions=None):
        super().__init__(calc_name, input_source_settings, max_nom_card_count, outlier_param, use_pop_nom_stats)
        self.partitions: List[List[UserPartitionInfo]] = partitions



"""todo"""
from typing import List

from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.idq.core.enums.data_entity_type import DataEntityType
from organon.idq.core.enums.dq_comparison_result_code import DqComparisonResultCode
from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.controls.base_dq_control import BaseDqControl
from organon.idq.domain.enums.dq_control_type import DqControlType
from organon.idq.domain.enums.dq_test_group_type import DqTestGroupType


class EmptyTableControl(BaseDqControl):
    """todo"""

    @classmethod
    def get_test_group_type(cls) -> DqTestGroupType:
        return DqTestGroupType.TABLE_SCHEMA_CONTROLS

    @staticmethod
    def get_control_type() -> DqControlType:
        return DqControlType.EMPTY_TABLE

    def get_description(self) -> str:
        return "Empty Table Control"

    def _execute_control(self) -> List[DqComparisonResult]:
        if self.comp_context.test_calc_result is None:
            raise ValueError("T-Calculation result is None.")
        if self.comp_context.test_calc_result.data_source_stats is None:
            raise KnownException("T-Calculation does not contain data source statistics, "
                                 "re-calculation may be required")
        if self.comp_context.test_calc_result.data_source_stats.row_count > 0:
            return []
        input_source = self.comp_context.test_calc_parameters.input_source_settings.source
        result = self._get_comparison_result(
            data_entity=DataEntityType.TABLE,
            data_entity_name=input_source.get_name(),
            result_code=DqComparisonResultCode.TABLE_IS_EMPTY,
            property_code="TableEmpty",
            message=f"Empty data in {input_source.get_name()}"
        )
        result.property_key_nominal = "True"
        return [result]



"""Includes HPOService class."""
from typing import Dict, List, Tuple, Union

import numpy as np
import pandas as pd
from sklearn.model_selection import GridSearchCV, PredefinedSplit, RandomizedSearchCV
from sklearn.model_selection._search import BaseSearchCV
from sklearn.pipeline import Pipeline
from skopt import BayesSearchCV

from organon.ml.modelling.algorithms.classifiers.rf_classifier import RFClassifier
from organon.ml.modelling.algorithms.core.abstractions.base_modeller import BaseModeller
from organon.ml.modelling.algorithms.core.enums.modeller import Modeller
from organon.ml.modelling.algorithms.services.algorithm_service import AlgorithmService

from organon.ml.modelling.model_selection.domain.objects.hpo_output import HPOOutput
from organon.ml.modelling.model_selection.domain.services.cross_validation_service import CrossValidationService
from organon.ml.modelling.model_selection.settings.enums.search_type import SearchType
from organon.ml.modelling.model_selection.settings.objects.hp_optimization_settings import \
    HPOptimizationSettings


class HPOService:
    """Hyperparameter optimization service"""

    @classmethod
    def execute(cls, settings: HPOptimizationSettings) -> HPOOutput:
        """Executes hyperparameter optimization with given settings and return HPOOutput"""
        is_default_search = settings.search_method is None
        if is_default_search:
            raise NotImplementedError  # analiz güncellenince eklenecek
        modellers = settings.modellers
        modeller_type = modellers[0].get_modeller_type()

        train_data = settings.train_data
        target_data = settings.target_data
        if settings.test_data is not None:
            train_data, target_data, predefined_split = cls._get_unified_data_and_split(
                train_data, target_data, settings.test_data, settings.test_target_data)
        else:
            predefined_split = CrossValidationService.get_predefined_split(settings.train_data,
                                                                           settings.target_data,
                                                                           modeller_type, cv_count=settings.cv_fold)

        modeller_instances = cls._get_modeller_instances(modellers)
        search_cv = cls._get_search_cv_instance(settings.search_method,
                                                modeller_instances, settings.modeller_params,
                                                predefined_split,
                                                settings.search_params, settings.scoring_metrics)
        search_cv.fit(train_data, target_data)

        output = cls._get_output(settings.train_data, settings.target_data, search_cv, modeller_instances)
        return output

    @classmethod
    def _get_search_cv_instance(cls, search_method: SearchType, modeller_instances, modeller_params,
                                predefined_split: PredefinedSplit,
                                search_params=None, scoring_metrics: List[str] = None) -> BaseSearchCV:
        cv_class = cls._validate_and_get_cv_class(search_method, search_params, scoring_metrics)
        params = cls._get_pipeline_params([tpl[1] for tpl in modeller_instances], modeller_params)

        pipeline = Pipeline([('mdl', RFClassifier())])  # placeholder classifier
        search_params = search_params if search_params is not None else {}
        scoring_metrics = scoring_metrics[0] \
            if scoring_metrics is not None and len(scoring_metrics) == 1 else scoring_metrics
        grid = cv_class(pipeline, params, cv=predefined_split, scoring=scoring_metrics,
                        refit=False, **search_params)
        return grid

    @classmethod
    def _get_output(cls, train_data: pd.DataFrame, target_data: pd.Series, search_cv: BaseSearchCV,
                    modeller_instances) -> HPOOutput:
        output = HPOOutput()
        output.best_modeller, output.best_params = cls._get_best_modeller_and_params(search_cv)
        output.best_modeller.set_params(**output.best_params)
        output.best_modeller.fit(train_data, target_data)
        output.summary = cls._get_metrics(search_cv, modeller_instances)
        return output

    @classmethod
    def _get_modeller_instances(cls, modellers: Union[List[Modeller]]):
        modeller_instances = [(modeller.name, AlgorithmService.get_modeller(modeller))
                              for modeller in modellers]
        return modeller_instances

    @classmethod
    def _get_best_modeller_and_params(cls, grid: BaseSearchCV):
        """return best modeller and its parameters."""
        # If there are multiple metrics, metrics are compared in order.
        # If there are two params with same score in first metric, next metric will be used for comparison
        results = grid.cv_results_
        scoring_metrics_list = ["score"] if not isinstance(grid.scoring, list) else grid.scoring

        all_mean_scores = []
        for i in range(len(results[f"mean_test_{scoring_metrics_list[0]}"])):
            param_mean_scores = []
            for metric in scoring_metrics_list:
                param_mean_scores.append(results[f"mean_test_{metric}"][i])
            all_mean_scores.append((i, param_mean_scores))
        all_mean_scores = sorted(all_mean_scores, key=lambda tpl: tpl[1], reverse=True)
        best_modeller, best_params = cls._get_params_with_actual_names(results["params"][all_mean_scores[0][0]])
        return best_modeller, best_params

    @classmethod
    def _get_metrics(cls, grid: BaseSearchCV, modeller_instances: List[Tuple[str, BaseModeller]]):
        scoring_metrics_list = ["score"] if not isinstance(grid.scoring, list) else grid.scoring
        cv_results = grid.cv_results_
        metrics = []
        num_splits = grid.n_splits_
        mean_metric_strings = []
        for i, params in enumerate(cv_results["params"]):
            modeller, params = cls._get_params_with_actual_names(params)
            res = {
                "modeller": next(tpl[0] for tpl in modeller_instances if tpl[1] == modeller),
                "params": params
            }
            for metric in scoring_metrics_list:
                mean_metric_str = metric
                if num_splits > 1:
                    mean_metric_str = f"mean {metric}"
                    for split_index in range(num_splits):
                        res[f"fold-{split_index} {metric}"] = cv_results[f"split{split_index}_test_{metric}"][i]
                res[mean_metric_str] = cv_results[f"mean_test_{metric}"][i]
                if mean_metric_str not in mean_metric_strings:
                    mean_metric_strings.append(mean_metric_str)
            metrics.append(res)
        frame = pd.DataFrame(metrics).sort_values(by=mean_metric_strings, ascending=False, ignore_index=True)

        return frame

    @classmethod
    def _get_params_with_actual_names(cls, grid_params_dict: dict) -> Tuple[BaseModeller, dict]:
        params = grid_params_dict.copy()
        best_modeller = params.pop("mdl")
        best_params = {param.split("mdl__")[1]: value for param, value in params.items()}
        return best_modeller, best_params

    @classmethod
    def _validate_and_get_cv_class(cls, search_method: SearchType, search_params: dict, scoring_metrics: List[str]):
        invalid_search_params = ["estimator", "cv", "scoring", "refit"]

        if search_method == SearchType.GRID:
            cv_class = GridSearchCV
            invalid_search_params.extend(["param_grid"])
        elif search_method == SearchType.RANDOM:
            cv_class = RandomizedSearchCV
            invalid_search_params.extend(["param_distributions"])
        else:
            cv_class = BayesSearchCV
            invalid_search_params.extend(["search_spaces"])

        if cv_class == BayesSearchCV and isinstance(scoring_metrics, list) and len(scoring_metrics) > 1:
            raise ValueError("Model Based Search cannot be run with multiple scoring metrics")

        if search_params is not None:
            invalid_entered_params = [param for param in search_params if param in invalid_search_params]
            if len(invalid_entered_params) > 0:
                raise ValueError(f"search_params cannot include these params: {','.join(invalid_entered_params)}")
        return cv_class

    @classmethod
    def _get_pipeline_params(cls, modeller_instances: List[BaseModeller],
                             modeller_params: List[Dict[str, list]]) -> List[Dict[str, list]]:
        all_params = []
        if modeller_params is None:
            modeller_params = [{} for _ in modeller_instances]
        for modeller, params_dict in zip(modeller_instances, modeller_params):
            new_params_dict = {"mdl": [modeller]}
            for param, list_of_values in params_dict.items():
                new_params_dict[f"mdl__{param}"] = list_of_values
            all_params.append(new_params_dict)
        return all_params

    @classmethod
    def _get_unified_data_and_split(cls, train_data: pd.DataFrame, train_target_data: pd.Series,
                                    test_data: pd.DataFrame,
                                    test_target_data: pd.Series) -> Tuple[pd.DataFrame, pd.Series,
                                                                          PredefinedSplit]:
        len_train = len(train_data)
        len_test = len(test_data)
        union_data = pd.concat([train_data, test_data], copy=False, ignore_index=True)
        union_target_data = pd.concat([train_target_data, test_target_data], copy=False, ignore_index=True)

        len_union = len_test + len_train
        test_fold = np.full(len_union, -1)
        test_fold[len_train:] = 1
        return union_data, union_target_data, PredefinedSplit(test_fold)



"""
This module includes enum class DateResolution.
"""
from enum import Enum


# pylint: disable=invalid-name
class DateResolution(Enum):
    """
    DateResolution
    """
    Year = 0
    Month = 1
    Day = 2
    Hour = 3
    Minute = 4
    Second = 5



"""Includes UnivariatePerformanceFeatureReductionOutput class"""
from typing import Dict

from organon.ml.feature_reduction.domain.objects.feature_reduction_output import FeatureReductionOutput


class UnivariatePerformanceFeatureReductionOutput(FeatureReductionOutput):
    """Output of univariate performance feature reduction"""

    def __init__(self):
        super().__init__()
        self.univariate_performance_result: Dict[str, float] = None



"""
This module is for unsupervised feature extraction
"""
from typing import List

from organon.afe.domain.modelling.base_modelling_service import BaseModellingService
from organon.afe.domain.modelling.businessobjects.data_frame_builder import DataFrameBuilder
from organon.afe.domain.modelling.unsupervised.afe_unsupervised_algorithm_settings import \
    AfeUnsupervisedAlgorithmSettings
from organon.afe.domain.modelling.unsupervised.unsupervised_modelling_helper import UnsupervisedModellingHelper
from organon.fl.core.businessobjects.dict_dataframe import DictDataFrame
from organon.fl.core.businessobjects.idata_partition import IDataPartition


class UnsupervisedFeatureExtractionService(BaseModellingService):
    """
    Class for Unsupervised Feature Extraction Service
    """

    def _get_selected_cols(self, num_threads: int,
                           partition: IDataPartition = None,
                           algorithm_settings: AfeUnsupervisedAlgorithmSettings = None,
                           frame_builder: DataFrameBuilder = None,
                           frame_with_all_columns: DictDataFrame = None,
                           is_final=False) -> List[str]:
        return UnsupervisedModellingHelper.get_selected_cols(num_threads, algorithm_settings, frame_with_all_columns)



"""Includes DataAugmentationSettings class."""
from organon.ml.common.enums.random_flip_type import RandomFlipType


class DataAugmentationSettings:
    """Settings for fine tuning in object classification"""

    def __init__(self, random_flip: RandomFlipType = RandomFlipType.HORIZONTAL_AND_VERTICAL,
                 random_rotation: float = 0.1, random_zoom: float = 0.1):
        self.random_flip: RandomFlipType = random_flip
        self.random_rotation: float = random_rotation
        self.random_zoom: float = random_zoom



"""Includes DqDfInputSourceSettings class."""

from organon.idq.domain.businessobjects.record_sources.dq_df_record_source import DqDfRecordSource
from organon.idq.domain.settings.abstractions.dq_base_input_source_settings import DqBaseInputSourceSettings


class DqDfInputSourceSettings(DqBaseInputSourceSettings[DqDfRecordSource]):
    """DQ input source settings for dataframe source type"""

    def __init__(self, source: DqDfRecordSource):
        super().__init__(source)



"""
This module includes MLApplicationOperations class.
"""
import threading

import organon
from organon.common.helpers import dev_mode_helper
from organon.fl.core.iocutil import ioc_helper
from organon.fl.core.iocutil.ioc_registration_item import IocRegistrationItem
from organon.fl.generic.interaction.fl_initializer import FlInitializer
from organon.ml.modelling.algorithms.classifiers.gam_classifier import GamClassifier
from organon.ml.modelling.algorithms.classifiers.gbm_classifier import GBMClassifier
from organon.ml.modelling.algorithms.classifiers.lightgbm_classifier import LightGBMClassifier
from organon.ml.modelling.algorithms.classifiers.logistic_regression_classifier import LogisticRegressionClassifier
from organon.ml.modelling.algorithms.classifiers.multi_layer_perceptron_classifier import MultiLayerPerceptronClassifier
from organon.ml.modelling.algorithms.classifiers.rf_classifier import RFClassifier
from organon.ml.modelling.algorithms.classifiers.xgboost_classifier import XGBoostClassifier
from organon.ml.modelling.algorithms.core.abstractions.base_regressor import BaseModeller
from organon.ml.modelling.algorithms.core.enums.modeller import Modeller
from organon.ml.modelling.algorithms.regressors.gam_regressor import GamRegressor
from organon.ml.modelling.algorithms.regressors.gbm_regressor import GBMRegressor
from organon.ml.modelling.algorithms.regressors.lasso_regressor import LassoRegressor
from organon.ml.modelling.algorithms.regressors.lightgbm_regressor import LightGBMRegressor
from organon.ml.modelling.algorithms.regressors.multi_layer_perceptron_regressor import MultiLayerPerceptronRegressor
from organon.ml.modelling.algorithms.regressors.rf_regressor import RFRegressor
from organon.ml.modelling.algorithms.regressors.ridge_regressor import RidgeRegressor
from organon.ml.modelling.algorithms.regressors.xgboost_regressor import XGBoostRegressor


class MLApplicationOperations:
    """
    MLApplicationOperations
    """

    APPLICATION_INITIALIZED = False
    __INITIALIZATION_LOCK = threading.Lock()

    @classmethod
    def initialize_app(cls):
        """Initializes application."""
        with MLApplicationOperations.__INITIALIZATION_LOCK:
            if not MLApplicationOperations.APPLICATION_INITIALIZED:
                cls._on_init()
                MLApplicationOperations.APPLICATION_INITIALIZED = True

    @classmethod
    def _on_init(cls):
        cls._initialize_fl()
        cls.register_types()

    @classmethod
    def _initialize_fl(cls):
        FlInitializer.application_initialize()

    @classmethod
    def init_dev_mode(cls, log_to_console: bool = True, log_file: str = "application.log"):
        """Initializes development mode."""
        dev_mode_helper.init_dev_mode(organon.ml.__name__, log_to_console=log_to_console, log_file=log_file)

    @classmethod
    def register_types(cls):
        """
        Registers ioc items
        """
        ioc_helper.register_type(IocRegistrationItem(Modeller.LOGISTIC_REGRESSION_CLASSIFIER.name,
                                                     BaseModeller, LogisticRegressionClassifier))

        ioc_helper.register_type(IocRegistrationItem(Modeller.GAM_CLASSIFIER.name,
                                                     BaseModeller, GamClassifier))

        ioc_helper.register_type(IocRegistrationItem(Modeller.GBM_CLASSIFIER.name,
                                                     BaseModeller, GBMClassifier))

        ioc_helper.register_type(IocRegistrationItem(Modeller.LIGHTGBM_CLASSIFIER.name,
                                                     BaseModeller, LightGBMClassifier))

        ioc_helper.register_type(IocRegistrationItem(Modeller.RF_CLASSIFIER.name,
                                                     BaseModeller, RFClassifier))

        ioc_helper.register_type(IocRegistrationItem(Modeller.XGBOOST_CLASSIFIER.name,
                                                     BaseModeller, XGBoostClassifier))

        ioc_helper.register_type(IocRegistrationItem(Modeller.MULTI_LAYER_PERCEPTRON_CLASSIFIER.name,
                                                     BaseModeller, MultiLayerPerceptronClassifier))

        ioc_helper.register_type(IocRegistrationItem(Modeller.GAM_REGRESSOR.name,
                                                     BaseModeller, GamRegressor))

        ioc_helper.register_type(IocRegistrationItem(Modeller.RIDGE.name,
                                                     BaseModeller, RidgeRegressor))

        ioc_helper.register_type(IocRegistrationItem(Modeller.LASSO.name,
                                                     BaseModeller, LassoRegressor))

        ioc_helper.register_type(IocRegistrationItem(Modeller.GBM_REGRESSOR.name,
                                                     BaseModeller, GBMRegressor))

        ioc_helper.register_type(IocRegistrationItem(Modeller.LIGHTGBM_REGRESSOR.name,
                                                     BaseModeller, LightGBMRegressor))

        ioc_helper.register_type(IocRegistrationItem(Modeller.RANDOM_FOREST_REGRESSOR.name,
                                                     BaseModeller, RFRegressor))

        ioc_helper.register_type(IocRegistrationItem(Modeller.XGBOOST_REGRESSOR.name,
                                                     BaseModeller, XGBoostRegressor))

        ioc_helper.register_type(IocRegistrationItem(Modeller.MULTI_LAYER_PERCEPTRON_REGRESSOR.name,
                                                     BaseModeller, MultiLayerPerceptronRegressor))



""" This module includes SimilarDistributionFeatureReduction"""
from typing import List, Dict, Tuple, Optional

import pandas as pd

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.fl.core.helpers.data_frame_helper import get_column_native_type
from organon.ml.feature_reduction.domain.enums.feature_reduction_types import FeatureReductionType
from organon.ml.feature_reduction.domain.helpers.univariate_performance_helper import \
    get_reduced_columns_via_performance
from organon.ml.feature_reduction.domain.objects.numeric_column_stats import NumericColumnStats
from organon.ml.feature_reduction.domain.objects.similar_dist_reduction_output import SimilarDistReductionOutput
from organon.ml.feature_reduction.domain.reductions.base_feature_reduction import BaseFeatureReduction
from organon.ml.feature_reduction.settings.objects.similar_distribution_feature_reduction_settings import \
    SimilarDistributionFeatureReductionSettings


class SimilarDistributionFeatureReduction(BaseFeatureReduction):
    """SimilarDistributionFeatureReduction class"""

    def _execute_reduction(self, settings: SimilarDistributionFeatureReductionSettings) -> SimilarDistReductionOutput:
        data = settings.data
        output = SimilarDistReductionOutput()
        output.feature_reduction_type = FeatureReductionType.SIMILAR_DISTRIBUTION
        included_columns = self._get_included_columns(data, settings.excluded_columns)
        numeric_columns = [col for col in included_columns if
                           (get_column_native_type(data, col) == ColumnNativeType.Numeric)
                           and (col != settings.target_column_name)
                           and (data[col].nunique() > settings.nunique_count)]
        if not numeric_columns:
            output.reduced_column_list = None
            return output
        numeric_columns_stats = self._get_stats(data, numeric_columns)
        similar_distribution_columns = self._get_similar_distribution_columns(numeric_columns_stats)
        reduced_columns, new_col_performances = self._find_reduced_columns(settings, similar_distribution_columns)
        output.reduced_column_list = reduced_columns
        output.new_univariate_performance_results = new_col_performances
        return output

    @staticmethod
    def _get_stats(data: pd.DataFrame, numeric_columns: List[str]) -> List[NumericColumnStats]:
        """
        Return numeric column stats which includes mean,std, percentile 25, percentile 50, percentile 75.
        Parameters
        ----------
        data
        numeric_columns

        Returns
        -------
        Return numeric columns stats list
        """

        stats_df = data[numeric_columns].describe()
        numeric_columns_stats = []
        for col in numeric_columns:
            numeric_column_stats = NumericColumnStats(col, stats_df[col]["mean"], stats_df[col]["std"],
                                                      stats_df[col]["25%"],
                                                      stats_df[col]["50%"],
                                                      stats_df[col]["75%"])
            numeric_columns_stats.append(numeric_column_stats)
        return numeric_columns_stats

    @staticmethod
    def _get_similar_distribution_columns(numeric_columns_stats: List[NumericColumnStats]) -> Dict[str, List[str]]:
        """
        Finds columns with similar distribution using numerical column statistics.        Parameters
        ----------
        numeric_columns_stats

        Returns
        -------
        Returns columns with the same distribution.
        """
        similar_distribution_columns = {}

        similar_distribution_columns_set = set()
        for count, value in enumerate(numeric_columns_stats):
            if value.column_name in similar_distribution_columns_set:
                continue
            column_list = []
            column_list.append(numeric_columns_stats[count].column_name)
            for j in range(count + 1, len(numeric_columns_stats)):

                if numeric_columns_stats[count] == numeric_columns_stats[j]:
                    column_list.append(numeric_columns_stats[j].column_name)
                    similar_distribution_columns_set.add(numeric_columns_stats[j].column_name)
            if len(column_list) > 1:
                similar_distribution_columns[numeric_columns_stats[count].column_name] = column_list
        return similar_distribution_columns

    @staticmethod
    def _find_reduced_columns(settings: SimilarDistributionFeatureReductionSettings,
                              similar_distribution_columns: Dict[str, List[str]]) -> \
            Tuple[List[str], Optional[Dict[str, float]]]:
        """
        Find reduced columns.
        Parameters
        ----------
        settings
        similar_distribution_columns
        numeric_columns_list

        Returns
        -------
        Returns columns list which reduced.
        """
        if len(similar_distribution_columns) == 0:
            return [], None
        reduced_columns = []
        if settings.target_column_name is not None and settings.target_type is not None:
            reduced_columns, new_col_performances = get_reduced_columns_via_performance(
                settings.data, settings.target_column_name, settings.target_type, settings.performance_metric,
                list(similar_distribution_columns.values()), settings.univariate_performance_result,
                random_state=settings.random_state)
            return reduced_columns, new_col_performances

        for column_list in similar_distribution_columns.values():
            column_list.sort()
            reduced_columns.extend(column_list[1:])
        return reduced_columns, None

    def get_description(self) -> str:
        return "Similar Distribution Feature Reduction"



"""Includes SamplingUserInputService class."""

from organon.ml.common.enums.target_type import TargetType
from organon.ml.common.helpers.user_input_service_helper import get_enum, get_default_if_none
from organon.ml.sampling.services.user_settings.user_sampling_settings import UserSamplingSettings
from organon.ml.sampling.settings.enums.sampling_strategy import SamplingStrategy
from organon.ml.sampling.settings.objects.sampling_settings import SamplingSettings


class SamplingUserInputService:
    """Service for validating user input and generating settings for sampling"""

    @classmethod
    def get_sampling_settings(cls, user_settings: UserSamplingSettings) -> SamplingSettings:
        """Validates settings entered by user and generates settings object"""
        target_type = get_enum(user_settings.target_type, TargetType)
        sampling_strategy = get_enum(user_settings.sampling_strategy, SamplingStrategy)
        strata_columns = get_default_if_none(user_settings.strata_columns, [])

        if user_settings.data is None:
            raise ValueError("Data to sample should be given")
        if user_settings.target_column_name is not None:
            if len(strata_columns) == 0:
                strata_columns.append(user_settings.target_column_name)
            if target_type is None:
                raise ValueError("Type of target column should be given.")

        if sampling_strategy in [SamplingStrategy.OVERSAMPLING, SamplingStrategy.UNDERSAMPLING]:
            if user_settings.target_column_name is None or \
                    target_type not in [TargetType.BINARY, TargetType.MULTICLASS]:
                raise ValueError("Oversampling and undersampling can only be used when there is a binary or multiclass "
                                 "target column")
            if target_type == TargetType.MULTICLASS and user_settings.sampling_ratio is not None:
                raise ValueError("Sampling ratio cannot be used for undersampling or oversampling with multiclass "
                                 "target")

        split_test_data = get_default_if_none(user_settings.split_test_data, True)
        test_split_ratio = get_default_if_none(user_settings.test_split_ratio, 0.3)
        sampling_ratio = get_default_if_none(user_settings.sampling_ratio, 0.5)

        return SamplingSettings(user_settings.data, user_settings.target_column_name, target_type, strata_columns,
                                split_test_data, test_split_ratio, sampling_strategy, sampling_ratio,
                                user_settings.data_sample_ratio)



"""
This module includes BinaryTarget class.
"""
from organon.afe.domain.common.reader_helper import get_values_from_kwargs


class BinaryTarget:
    """
    Information for AFE target columns of 'Binary' type
    """

    ATTR_DICT = {
        "positive_category": str,
        "negative_category": str,
        "indeterminate_category": str,
        "exclusion_category": str
    }

    def __init__(self, **kwargs):
        self.positive_category: str = None
        self.negative_category: str = None
        self.indeterminate_category: str = None
        self.exclusion_category: str = None

        get_values_from_kwargs(self, self.ATTR_DICT, kwargs)

    def __deepcopy__(self, memo):
        output = BinaryTarget()
        output.positive_category = self.positive_category
        output.negative_category = self.negative_category
        output.indeterminate_category = self.indeterminate_category
        output.exclusion_category = self.exclusion_category
        return output



"""Includes MatplotlibCommonFunctions class."""

from datetime import timedelta
from typing import TypeVar, Type

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from matplotlib.axes import Axes, BarContainer
from matplotlib.lines import Line2D
from matplotlib.ticker import MaxNLocator

from organon.afe.domain.reporting import dashboard_helper
from organon.afe.domain.reporting.abstract_report_helper import AbstractReportHelper
from organon.afe.domain.reporting.base_afe_output_report import BaseAfeOutputReport
from organon.afe.domain.reporting.dashboard_helper import TablePagination
# ax kısaltması bir convention
# pylint: disable=invalid-name
from organon.fl.core.exceptionhandling.known_exception import KnownException

ReportHelperType = TypeVar("ReportHelperType", bound=AbstractReportHelper)


class MatplotlibCommonFunctions:
    """Common functions for matplotlib report helper classes."""

    @classmethod
    def generate_afe_report_dashboard(cls, report_helper_type: Type[ReportHelperType], afe_report: BaseAfeOutputReport):
        """Creates afe report dashboard"""
        fig, ax1 = plt.subplots()
        fig.canvas.manager.set_window_title("Output Features")
        ax1.set_title("Output Features")
        ax1.set_axis_off()
        table = cls.__get_afe_feature_table(report_helper_type, afe_report, ax1, rows_per_page=3)
        table.add_buttons(fig)
        plt.show()
        fig, ax2 = plt.subplots()
        fig.canvas.manager.set_window_title("Resource Usage")
        cls.__get_resource_usage_ax(afe_report, ax2)
        plt.show()

    @classmethod
    def generate_afe_feature_dashboard(cls, report_helper_type: Type[ReportHelperType],
                                       afe_report: BaseAfeOutputReport):
        """Creates a table including afe output features information"""

        fig, ax = plt.subplots()
        ax.set_axis_off()
        fig.patch.set_visible(False)
        fig.canvas.manager.set_window_title("Afe Feature Dashboard")
        ax.set_title("Output Features")

        table = cls.__get_afe_feature_table(report_helper_type, afe_report, ax, rows_per_page=10)
        table.add_buttons(fig)

        plt.show()

        return ax

    @classmethod
    def generate_feature_extended_dashboard(cls, afe_report: BaseAfeOutputReport):
        """Creates extended feature report"""

        quantity_summary = cls.__convert_to_df(
            afe_report.feature_counts_report.by_quantity, "Quantity")
        dimension_summary = cls.__convert_to_df(afe_report.feature_counts_report.by_dimension,
                                                "Dimension")
        operator_summary = cls.__convert_to_df(afe_report.feature_counts_report.by_operator,
                                               "Operator")
        tw_summary = cls.__convert_to_df(afe_report.feature_counts_report.by_time_window,
                                         "Time Window")
        total_feature = len(afe_report.output_features)
        rc2 = {"font.size": 8}
        with plt.rc_context(rc2):
            fig, axs = plt.subplots(nrows=2, ncols=2, figsize=(30, 20), constrained_layout=True)
            for ax in axs.reshape(-1):
                ax.xaxis.set_major_locator(MaxNLocator(integer=True))

            sns.barplot(x="Count", y="Quantity", data=quantity_summary, orient="h", ax=axs[0, 0]).set(
                title='Quantity Columns Distribution', ylabel=None)
            cls.__arrange_y_tick_labels(axs[0, 0], quantity_summary["Quantity"])
            sns.barplot(x="Count", y="Dimension", data=dimension_summary, orient="h", ax=axs[0, 1]).set(
                title='Dimension Columns Distribution', ylabel=None)
            cls.__arrange_y_tick_labels(axs[0, 1], dimension_summary["Dimension"])
            sns.barplot(x="Count", y="Operator", data=operator_summary, orient="h", ax=axs[1, 0]).set(
                title='Operators Distribution', ylabel=None)
            cls.__arrange_y_tick_labels(axs[1, 0], operator_summary["Operator"])
            sns.barplot(x="Count", y="Time Window", data=tw_summary, orient="h", ax=axs[1, 1],
                        order=tw_summary['Time Window']).set(title='Time Window Distribution', ylabel=None)
            cls.__arrange_y_tick_labels(axs[1, 1], tw_summary["Time Window"])
            fig.suptitle(f'Total Number of Features: {total_feature}', fontsize=20)
            plt.show()

    @classmethod
    def __arrange_y_tick_labels(cls, ax, tick_labels):
        y_tick_lbl_size = 8 if len(tick_labels) <= 7 else 64 / len(tick_labels)
        max_label_len = max([len(str(label)) for label in tick_labels])
        ax.tick_params(labelsize=y_tick_lbl_size)
        dashboard_helper.wrap_labels(ax, int(2000 / (y_tick_lbl_size * max_label_len)), True, axis="y")

    @classmethod
    def generate_feature_importance_dashboard(cls, afe_report: BaseAfeOutputReport, target_index: int = None,
                                              first_x_feature: int = None):
        """Creates afe feature importance dashboard"""
        if afe_report.final_column_metrics is None:
            raise ValueError("afe report does not include final column metrics")
        metrics_to_report = list(enumerate(afe_report.final_column_metrics))
        if target_index is not None:
            if len(afe_report.final_column_metrics) <= target_index:
                raise KnownException(f"Target with index {target_index} does not exist. "
                                     f"Num targets: {len(afe_report.final_column_metrics)}")
            metrics_to_report = [(target_index, afe_report.final_column_metrics[target_index])]
        for i, metrics in metrics_to_report:
            first_x_feature = first_x_feature if first_x_feature is not None else len(metrics["feature_importances"])
            plt.figure(figsize=(30, 20))
            data = metrics["feature_importances"].sort_values(by="Value", ascending=False)[:first_x_feature]
            sns.barplot(x="Value", y="Feature", data=data, orient="h", order=data["Feature"])
            ax = plt.gca()
            y_tick_lbl_size = 8 if len(data) <= 10 else 90 / len(data)
            ax.tick_params(labelsize=y_tick_lbl_size)
            dashboard_helper.wrap_labels(ax, int(160 / y_tick_lbl_size), True, axis="y")
            plt.title(f"Feature Importances (Target {i + 1})")
            plt.show()

    @classmethod
    def generate_roc_curve_dashboard(cls, afe_report: BaseAfeOutputReport, target_index: int = None, target_class=None):
        """Creates roc curve plot for features in given target"""

        if afe_report.final_column_metrics is None:
            raise ValueError("afe report does not include final column metrics")
        metrics_to_report = list(enumerate(afe_report.final_column_metrics))
        if target_index is not None:
            if len(afe_report.final_column_metrics) <= target_index:
                raise KnownException(f"Target with index {target_index} does not exist. "
                                     f"Num targets: {len(afe_report.final_column_metrics)}")
            metrics_to_report = [(target_index, afe_report.final_column_metrics[target_index])]

        for i, metrics in metrics_to_report:
            if "fpr" not in metrics or "tpr" not in metrics or "auc" not in metrics:
                raise ValueError(f"metrics for target {i} does not include all of required values: (fpr,tpr,auc)")
            fpr_dict, tpr_dict = metrics["fpr"].copy(), metrics["tpr"].copy()
            auc_val_dict = metrics["auc"].copy()
            if len(fpr_dict) > 0 and len(tpr_dict) > 0 and len(auc_val_dict) > 0:
                if "all" in fpr_dict:
                    plt.figure()
                    fpr, tpr, auc_val = fpr_dict["all"], tpr_dict["all"], auc_val_dict["all"]
                    plt.title(f"Receiver Operating Characteristic (Target {i + 1})")
                    plt.plot(fpr, tpr, "b", label=f"AUC = {auc_val:0.2F}")
                    plt.legend(loc="lower right")
                    plt.plot([0, 1], [0, 1], "k--")
                    plt.xlim([0, 1])
                    plt.ylim([0, 1])
                    plt.ylabel("True Positive Rate")
                    plt.xlabel("False Positive Rate")
                    plt.show()
                elif "micro" in fpr_dict:
                    micro_fpr, macro_fpr = fpr_dict.pop("micro"), fpr_dict.pop("macro")
                    micro_tpr, macro_tpr = tpr_dict.pop("micro"), tpr_dict.pop("macro")
                    micro_auc, macro_auc = auc_val_dict.pop("micro"), auc_val_dict.pop("macro")

                    classes = list(fpr_dict.keys())
                    classes_to_report = classes
                    if target_class is not None:
                        if target_class not in classes:
                            raise ValueError(f"'{target_class}' is not found in target classes")
                        classes_to_report = [target_class]
                    for _cls in classes_to_report:
                        plt.figure()
                        plt.title(f"ROC Class of '{_cls}' (Target {i + 1})")
                        plt.plot(micro_fpr, micro_tpr, label=f"micro-average ROC curve (area={micro_auc:0.2F})",
                                 color="deeppink", linestyle=":", linewidth=4)
                        plt.plot(macro_fpr, macro_tpr, "r",
                                 label=f"macro-average ROC curve (area={macro_auc:0.2F})",
                                 color="navy", linestyle=":", linewidth=4)
                        plt.plot(fpr_dict[_cls], tpr_dict[_cls], color="orange",
                                 label=f"ROC curve of class '{_cls}' (area={auc_val_dict[_cls]:0.2F})")
                        plt.legend(loc="lower right")
                        plt.plot([0, 1], [0, 1], "k--")
                        plt.xlim([0, 1])
                        plt.ylim([0, 1])
                        plt.ylabel("True Positive Rate")
                        plt.xlabel("False Positive Rate")
                        plt.show()

    @classmethod
    def generate_coarse_class_dashboard(cls, afe_report: BaseAfeOutputReport, feature_name: str):
        """Creates bar plot showing transformation map corresponding to given feature"""
        fig, ax = plt.subplots()
        fig.canvas.manager.set_window_title("Coarse Class Dashboard")
        cls.__get_coarse_class_bar_chart(afe_report, feature_name, ax)
        ax.set_xlabel("Range")
        ax.set_ylabel("Value")
        ax.set_title("Transformation Map")
        fig.autofmt_xdate()

        # PLOT LINE ON BAR PLOT WITH MATPLOTLIB (ileride kullanılacak)
        # pylint: disable=pointless-string-statement
        """
        _map = afe_report.transformations[feature_name].tabulate_as_text()
        map_vals = [float(i) for i in list(_map.values())]
        keys = list(_map.keys())
        ax2 = ax.twinx()
        ax2.plot(keys,[i for i,v in enumerate(keys)],label="target", color="orange")
        plt.legend()
        """

        fig.tight_layout()
        plt.show()
        return fig

    @classmethod
    def generate_memory_usage_dashboard(cls, afe_report: BaseAfeOutputReport):
        """Creates a line-scatter plot of memory usage by date"""
        fig, ax = plt.subplots()
        fig.canvas.manager.set_window_title("Memory Usage Dashboard")
        ax.set_title("Memory Usage")
        cls.__get_memory_usage_scatter(afe_report, ax)
        cls.__add_event_lines_to_ax(afe_report, ax)
        fig.autofmt_xdate()
        dashboard_helper.set_date_ticker(ax)
        plt.show()

    @classmethod
    def generate_cpu_usage_dashboard(cls, afe_report: BaseAfeOutputReport):
        """Creates a line-scatter plot of cpu usage by date"""
        fig, ax = plt.subplots()
        fig.canvas.manager.set_window_title("CPU Usage Dashboard")
        ax.set_title("CPU Usage")
        cls.__get_cpu_usage_scatter(afe_report, ax)
        cls.__add_event_lines_to_ax(afe_report, ax, for_cpu=True)
        # fig = create_figure([scatter], "CPU Usage", "Date", "Usage")
        # fig.show()
        fig.autofmt_xdate()
        dashboard_helper.set_date_ticker(ax)
        plt.show()

    @classmethod
    def __get_resource_usage_ax(cls, afe_report, ax):
        line1 = cls.__get_memory_usage_scatter(afe_report, ax)[0]

        color = "tab:blue"
        ax.set_ylabel("Memory Usage(MB)", color=color)
        ax.tick_params(axis='y', labelcolor=color)

        ax2 = ax.twinx()
        color2 = "tab:red"
        ax2.set_ylabel("CPU Usage(percentage)", color=color2)
        ax2.tick_params(axis='y', labelcolor=color2)
        line2 = cls.__get_cpu_usage_scatter(afe_report, ax2, color=color2)[0]
        cls.__add_event_lines_to_ax(afe_report, ax)
        ax.set_title("Resource Usage")
        return [line1, line2]

    @classmethod
    def generate_resource_usage_dashboard(cls, afe_report: BaseAfeOutputReport):
        """Creates a line-scatter plots of memory and cpu usages by date"""
        fig, ax = plt.subplots()
        lines = cls.__get_resource_usage_ax(afe_report, ax)
        fig.canvas.manager.set_window_title('Resource Usage Dashboard')
        fig.autofmt_xdate()
        dashboard_helper.set_date_ticker(ax)
        plt.legend(lines, [line.get_label() for line in lines])
        plt.show()

    @classmethod
    def __get_coarse_class_bar_chart(cls, afe_report: BaseAfeOutputReport, feature_name: str, ax: Axes) -> BarContainer:
        """Returns a matplotlib BarContainer object generated for afe transformation map of given feature"""
        _map = afe_report.transformations[feature_name].tabulate_as_text()
        limits = list(_map.keys())
        map_vals = [float(i) for i in list(_map.values())]

        bar_chart = dashboard_helper.get_bar_chart(limits, map_vals, "Observation Ratio", ax)

        return bar_chart

    @classmethod
    def __get_afe_feature_table(cls, report_helper_type: Type[ReportHelperType],
                                afe_report: BaseAfeOutputReport, ax: Axes, rows_per_page=10) -> TablePagination:
        """Returns a paginated table storing afe output features information"""
        data_frame = report_helper_type.get_output_features_as_data_frame(afe_report)
        table = dashboard_helper.get_data_table_plot(data_frame, ax, rows_per_page=rows_per_page)

        return table

    @classmethod
    def __get_memory_usage_scatter(cls, afe_report: BaseAfeOutputReport, ax: Axes) -> Line2D:
        """Returns a matplotlib Line2D object generated for memory usage by date"""
        usage = afe_report.memory_usage
        return dashboard_helper.get_line_scatter_plot(list(usage.keys()), list(usage.values()), "Memory Usage", ax)

    @classmethod
    def __get_cpu_usage_scatter(cls, afe_report: BaseAfeOutputReport, ax: Axes, color=None) -> Line2D:
        """Returns a matplotlib Line2D object generated for cpu usage by date"""
        usage = afe_report.cpu_usage
        return dashboard_helper.get_line_scatter_plot(list(usage.keys()), list(usage.values()), "CPU Usage", ax,
                                                      color=color)

    @classmethod
    def __add_event_lines_to_ax(cls, afe_report, ax: Axes, for_cpu=False):
        """Adds vertical lines for events in plot"""
        tick_vals = list(afe_report.event_times.values())
        tick_texts = list(afe_report.event_times.keys())
        values_dict = afe_report.memory_usage
        if for_cpu:
            values_dict = afe_report.cpu_usage
        max_val = max(values_dict.values())
        min_val = min(values_dict.values())
        total_time = max(values_dict.keys()) - min(values_dict.keys())
        for i, val in enumerate(tick_vals):
            if i < len(tick_vals) - 1:
                if (tick_vals[i + 1] - val).total_seconds() < total_time.total_seconds() / 100:
                    # prevents annotation overlapping by eliminating steps with a low execution time
                    continue
            ax.axvline(val, color="green", linestyle="--")
            ax.text(val + timedelta(seconds=0.1), (min_val + max_val) / 2, tick_texts[i], rotation=90,
                    color="green",
                    fontweight="bold", zorder=10, va="center")

    @classmethod
    def __convert_to_df(cls, count_dict: dict, property_name: str):
        new_dict = {property_name: [], "Count": []}
        for key, value in count_dict.items():
            new_dict[property_name].append(key)
            new_dict["Count"].append(value)
        return pd.DataFrame(new_dict)



"""
This module includes enum class AfeDateColumnType.
"""
from enum import Enum


# pylint: disable=invalid-name
class AfeDateColumnType(Enum):
    """
    AfeDateColumnType
    """
    YyyyMmDd = 0
    YyyyMm = 1
    DateTime = 2
    CustomFormat = 3



"""Includes DqDbFullProcessInput class."""
from organon.idq.domain.settings.abstractions.dq_full_process_input import DqFullProcessInput
from organon.idq.domain.settings.calculation.dq_df_calculation_parameters import DqDfCalculationParameters
from organon.idq.domain.settings.comparison.dq_df_comparison_parameters import DqDfComparisonParameters


class DqDfFullProcessInput(DqFullProcessInput[DqDfCalculationParameters, DqDfComparisonParameters]):
    """DQ comparison parameters for pandas dataframe source type"""



"""Includes BaseModellingService class"""
import abc
from typing import List

from organon.afe.core.businessobjects.afe_static_objects import AfeStaticObjects
from organon.afe.dataaccess.services.i_sample_data_service import ISampleDataService
from organon.afe.domain.modelling.abstract_modelling_service import AbstractModellingService
from organon.afe.domain.modelling.auto_column_type_decider_service import AutoColumnTypeDeciderService
from organon.afe.domain.modelling.businessobjects.afe_column import AfeColumn
from organon.afe.domain.modelling.businessobjects.afe_feature import AfeFeature
from organon.afe.domain.modelling.businessobjects.data_frame_builder import DataFrameBuilder
from organon.afe.domain.modelling.businessobjects.multi_target_entity_container import MultiTargetEntityContainer
from organon.afe.domain.modelling.businessobjects.target_file_record_collection import TargetFileRecordCollection
from organon.afe.domain.modelling.businessobjects.transaction_file_record_collection import \
    TransactionFileRecordCollection
from organon.afe.domain.reporting.afe_model_output import AfeModelOutput
from organon.afe.domain.settings.afe_modelling_settings import AfeModellingSettings
from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.fl.core.helpers import list_helper
from organon.fl.logging.helpers.log_helper import LogHelper


class BaseModellingService(AbstractModellingService[AfeModellingSettings, AfeModelOutput, AfeColumn],
                           metaclass=abc.ABCMeta):
    """Base class for supervised and unsupervised modelling"""

    def __init__(self, settings: AfeModellingSettings):
        super().__init__(settings)

    @classmethod
    def _get_model_output_instance(cls) -> AfeModelOutput:
        return AfeModelOutput()

    def _control_date_column_constraint(self):
        feature_gen_settings = self._settings.data_source_settings.trx_descriptor.feature_gen_setting
        no_date_in_target_settings = self._settings.data_source_settings.target_descriptor.date_column is None
        if feature_gen_settings.date_column is not None and no_date_in_target_settings:
            raise KnownException("Since feature generation settings contain a date column, "
                                 "the target definition must have a date column")

    def _decide_dimension_and_quantity_columns(self, sample_service: ISampleDataService):
        data_source_settings = self._settings.data_source_settings
        trx_descriptor = data_source_settings.trx_descriptor

        auto_column_decider = AutoColumnTypeDeciderService(sample_service,
                                                           [trx_descriptor.feature_gen_setting],
                                                           trx_descriptor.entity_column_name,
                                                           data_source_settings.auto_column_decider_settings)

        d_columns_per_setting, q_columns_per_setting = auto_column_decider.decide_quantity_dimension_columns()
        feature_generation_settings = trx_descriptor.feature_gen_setting
        d_columns = d_columns_per_setting[0]
        q_columns = q_columns_per_setting[0]

        feature_generation_settings.dimension_columns = \
            self.__update_column_list(d_columns, AfeStaticObjects.empty_dimension_column)
        feature_generation_settings.quantity_columns = \
            self.__update_column_list(q_columns, AfeStaticObjects.empty_quantity_column)
        if feature_generation_settings.date_column is not None:
            date_col_name = feature_generation_settings.date_column.column_name
        else:
            date_col_name = " No_Date_Col - 0"
        LogHelper.info(f'Decided columns for {date_col_name} '
                       f'are \nDimension Columns:\t{feature_generation_settings.dimension_columns}'
                       f'\nQuantity Columns:\t{feature_generation_settings.quantity_columns}')

    def _get_non_empty_date_columns_dict(self):
        date_col = self._settings.data_source_settings.trx_descriptor.feature_gen_setting.date_column
        if date_col is not None:
            return {date_col.column_name: date_col}
        return {}

    def _get_interval_per_entity_per_date_col(self, multi_target_entity_container: MultiTargetEntityContainer):
        interval_per_entity_per_date_col = {}
        feature_generation_settings = self._settings.data_source_settings.trx_descriptor.feature_gen_setting

        if feature_generation_settings.date_column is not None:
            interval_per_entity = multi_target_entity_container.get_date_interval_per_entity(
                feature_generation_settings.date_resolution, max(feature_generation_settings.horizon_list),
                feature_generation_settings.date_offset)
            interval_per_entity_per_date_col[feature_generation_settings.date_column.column_name] \
                = interval_per_entity
        return interval_per_entity_per_date_col

    def _get_output_features(self, reduced_columns_list: List[AfeColumn]):
        output_features = {}
        for i, col in enumerate(reduced_columns_list):
            feature_name = f"{self._settings.output_settings.feature_name_prefix}_{i}"
            feature = AfeFeature(col)
            feature.feature_name = feature_name
            output_features[feature_name] = feature
        return output_features

    def _get_data_frame_builder(self, target_file_record_collection: TargetFileRecordCollection,
                                trx_file_record_collection: TransactionFileRecordCollection):
        all_dimensions, all_horizons = {}, {}
        data_source_settings = self._settings.data_source_settings
        feature_generation_settings = data_source_settings.trx_descriptor.feature_gen_setting
        all_dimensions[0] = set(feature_generation_settings.dimension_columns)
        all_horizons[0] = set(feature_generation_settings.horizon_list)

        frame_builder = DataFrameBuilder(data_source_settings.target_descriptor,
                                         target_file_record_collection,
                                         trx_file_record_collection,
                                         dimensions_per_date_col=all_dimensions,
                                         horizons_per_date_col=all_horizons)
        return frame_builder

    def _reduce_for_settings(self, frame_builder, all_columns_frame, reduction_settings, file):
        trx_descriptor = self._settings.data_source_settings.trx_descriptor
        partition = self._reduce_for_feature_gen_setting(frame_builder, 0,
                                                         trx_descriptor.feature_gen_setting, reduction_settings,
                                                         all_columns_frame, None, file)
        return partition

    @staticmethod
    def __update_column_list(base_list: List[str], new_column: str) -> List[str]:
        if list_helper.is_null_or_empty(base_list):
            return [new_column]
        base_list.append(new_column)
        return list(set(base_list))



"""Includes PsiNumericSignalControl class."""
from typing import List, Dict

from organon.fl.core.collections.sorted_dict import SortedDict
from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.fl.logging.helpers.log_helper import LogHelper
from organon.fl.mathematics.businessobjects.sets.interval import Interval
from organon.idq.core.dq_constants import DqConstants
from organon.idq.core.enums.data_entity_type import DataEntityType
from organon.idq.core.enums.dq_comparison_result_code import DqComparisonResultCode
from organon.idq.domain.businessobjects.dq_calculation_result import DqCalculationResult
from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.businessobjects.statistics.dq_sample_numerical_statistics import DqSampleNumericalStatistics
from organon.idq.domain.controls.base_psi_signal_control import BasePsiSignalControl
from organon.idq.domain.enums.dq_control_type import DqControlType
from organon.idq.domain.enums.signal_type import SignalType
from organon.idq.domain.helpers.statistics.dq_statistical_functions import DqStatisticalFunctions


class PsiNumericSignalControl(BasePsiSignalControl):
    """Control class psi comparison for numeric column"""

    @staticmethod
    def get_control_type() -> DqControlType:
        return DqControlType.PSI_NUMERIC_VALUES

    def get_description(self) -> str:
        return "Psi Numeric Signals"

    def _execute_control(self) -> List[DqComparisonResult]:
        if self.comp_context.test_calc_parameters.is_existing_calculation and \
                self.comp_context.test_calc_result.sample_data is None:
            LogHelper.warning("Will not execute PSI control for numeric columns since test calculation "
                              "is given as existing calculation with dataframe source")
            return []
        numeric_comp_cols = self.comp_context.get_comparison_cols_by_native_type(ColumnNativeType.Numeric)
        sample_data = self.comp_context.test_calc_result.sample_data
        results = []
        test_group_info = self.comp_context.test_group_info[self.get_test_group_type()]
        for col in numeric_comp_cols:
            col_name = col.column_name
            filtered_controls = self._get_filtered_columns(col, test_group_info)
            if len(filtered_controls) > 0 and col_name in sample_data.data_frame.columns:
                test_array = sample_data.data_frame[col_name]
                self.__control_executor(test_array, filtered_controls, col_name, results)

        return results

    def __control_executor(self, test_array,
                           filtered_controls: List[DqCalculationResult], col_name: str,
                           results: List[DqComparisonResult]):
        max_psi = 0.0
        maximum_psi_sample, maximum_psi_category = "", ""
        maximum_psi_control_frequencies: SortedDict[str, float] = SortedDict()
        maximum_psi_test_frequencies: SortedDict[str, float] = SortedDict()
        default_value_string = "A000001"
        non_default_value_string = "B000001"
        psi_per_control_set: Dict[str, float] = {}
        for control in filtered_controls:
            stat_collection = control.sample_stats.numerical_statistics
            if col_name not in stat_collection:
                continue
            control_stats = stat_collection[col_name]
            control_frequencies = self._get_control_frequencies(control_stats, default_value_string,
                                                                non_default_value_string)
            test_frequencies = self._get_test_frequencies(control_stats, default_value_string,
                                                          non_default_value_string, test_array)
            psi_tuple = DqStatisticalFunctions.population_stability_index(control_frequencies,
                                                                          test_frequencies,
                                                                          DqConstants.MINIMUM_INJECTION)
            psi_value = psi_tuple[1]
            sample_to_string = control.calculation_name
            psi_per_control_set[sample_to_string] = psi_value
            if psi_value > max_psi:
                max_psi = psi_value
                maximum_psi_sample = sample_to_string
                maximum_psi_category = psi_tuple[0]
                maximum_psi_control_frequencies = control_frequencies
                maximum_psi_test_frequencies = test_frequencies
        psi_signal = self.get_signal_type(max_psi)
        if psi_signal == SignalType.GREEN:
            return
        result_code = DqComparisonResultCode.PSI_RED_SIGNAL if psi_signal == SignalType.RED \
            else DqComparisonResultCode.PSI_YELLOW_SIGNAL
        self.__check_psi_signal(col_name, result_code, results, psi_signal, "PsiSignal")
        self.check_maximum_psi(col_name, result_code, results, max_psi, "MaximumPsi")
        self.check_maximum_psi_category(col_name, result_code, results, maximum_psi_category,
                                        "MaximumPsiCategory")
        self.check_maximum_psi_category(col_name, result_code, results, maximum_psi_category,
                                        "MaximumPsiInterval")
        self.check_maximum_psi_control_set(col_name, result_code, results, maximum_psi_sample,
                                           "MaximumPsiControlSet")
        self.check_test_frequency(col_name, result_code, results, maximum_psi_test_frequencies,
                                  "NumericalColumnTestFrequencies")
        self.check_max_psi_control_frequency(col_name, result_code, results,
                                             maximum_psi_control_frequencies,
                                             "NumericalColumnMaximumPsiControlFrequencies")
        self.check_psi_per_control_set(col_name, result_code, results, psi_per_control_set,
                                       "PsiPerControlSet")

    def __check_psi_signal(self, col_name: str, result_code: DqComparisonResultCode,
                           results: List[DqComparisonResult], psi_signal: SignalType, property_key: str):
        result = DqComparisonResult(
            data_entity=DataEntityType.COLUMN,
            data_entity_name=col_name,
            test_group=self.get_test_group_type(),
            result_code=result_code,
            property_code=property_key,
            message=f"Distribution is unstable for the numeric column {col_name}")
        result.property_key_nominal = psi_signal.name
        results.append(result)

    @staticmethod
    def _get_control_frequencies(control_stats: DqSampleNumericalStatistics,
                                 default_value_string: str, non_default_value_string: str) -> SortedDict[str, float]:

        default_value_frequencies = control_stats.missing_values_frequencies
        default_values_count = 0.0
        if default_value_frequencies is not None:
            default_values_count = sum(default_value_frequencies.values())
        result: SortedDict[str, float] = SortedDict()
        result[default_value_string] = default_values_count
        interval_statistics = control_stats.interval_statistics
        if interval_statistics is None:
            result[non_default_value_string] = 0
        else:
            for interval_statistic_key, interval_statistic_value in interval_statistics.items():
                interval_frequency = interval_statistic_value.count
                result[interval_statistic_key.to_string()] = interval_frequency
        return result

    @staticmethod
    def _get_test_frequencies(control_stats: DqSampleNumericalStatistics,
                              default_value_string: str, non_default_value_string: str, test_array) \
            -> SortedDict[str, float]:
        interval_statistics = control_stats.interval_statistics
        upper_bounds_map: SortedDict[float, Interval] = SortedDict()
        test_normal_values: SortedDict[float, float] = SortedDict()
        if interval_statistics is not None:
            for interval_statistic_key in interval_statistics:
                upper_bounds_map[interval_statistic_key.upper_bound] = interval_statistic_key
                test_normal_values[interval_statistic_key.upper_bound] = 0.0

        default_values = control_stats.missing_values if control_stats.missing_values is not None else []
        default_value_count = 0.0
        for value in test_array:
            if value in default_values:
                default_value_count = default_value_count + 1
            else:
                for upper_bound_key in upper_bounds_map:
                    if value <= upper_bound_key:
                        test_normal_values[upper_bound_key] = test_normal_values[upper_bound_key] + 1
                        break

        result: SortedDict[str, float] = SortedDict()
        result[default_value_string] = default_value_count
        if interval_statistics is None:
            result[non_default_value_string] = len(test_array) - default_value_count
        else:
            for test_normal_values_key, test_normal_values_value in test_normal_values.items():
                result[(upper_bounds_map[test_normal_values_key]).to_string()] = test_normal_values_value
        return result



"""This module includes DictDataFrameOperations class"""
from typing import List, Dict, Any

from organon.fl.core.businessobjects.dict_dataframe import DictDataFrame
from organon.fl.core.businessobjects.idataframe import IDataFrame
from organon.fl.mathematics.helpers.idataframe_operations import IDataFrameOperations


class DictDataFrameOperations(IDataFrameOperations):
    """Class for Dictionary Dataframe Operations."""

    @staticmethod
    def get_min_value(df_obj: IDataFrame, col_name: str, skipna=True) -> float:
        pass

    @staticmethod
    def get_max_value(df_obj: IDataFrame, col_name: str, skipna=True) -> float:
        pass

    @staticmethod
    def get_mean_value(df_obj: IDataFrame, col_name: str, skipna=True) -> float:
        pass

    @staticmethod
    def get_trimmed_mean_value(df_obj: IDataFrame, col_name: str, lower_bound: float, upper_bound: float) -> float:
        pass

    @staticmethod
    def get_variance_value(df_obj: IDataFrame, col_name: str, skipna=True) -> float:
        pass

    @staticmethod
    def get_sum_value(df_obj: IDataFrame, col_name: str, skipna=True) -> float:
        pass

    @staticmethod
    def get_percentile_values(df_obj: IDataFrame, col_name: str, percentiles: List[float]) -> Dict[float, float]:
        pass

    @staticmethod
    def get_frequencies(df_obj: IDataFrame, col_name: str, values: list = None) -> Dict[Any, int]:
        pass

    @staticmethod
    def get_col_df_without_values(df_obj: IDataFrame, col_name: str, values: list) -> IDataFrame:
        pass

    @staticmethod
    def get_size(df_obj: IDataFrame):
        pass

    @staticmethod
    def get_col_values(df_obj: IDataFrame, col_name: str) -> list:
        pass

    @staticmethod
    def get_column_stability(df_obj: IDataFrame):
        """:returns boolean df depends on whether a column is stable or not"""
        raise NotImplementedError

    @staticmethod
    def join(df1: DictDataFrame, df2: DictDataFrame):
        """:returns pandas join function equivalent for dict"""
        raise NotImplementedError

    @staticmethod
    def get_column_distinct_counts(df_obj: DictDataFrame, col_names: List[str] = None):
        """:returns number of unique values per column"""
        raise NotImplementedError



"""
This module keeps the helper functions for object type.
"""
import codecs
import inspect
import pickle  # nosec
import zlib
from typing import Tuple

from organon.fl.security.helpers.encryption_helper import EncryptionHelper


class DecompressionError(Exception):
    """Error raised when decompression fails"""


def get_encrypted_binary(obj, secret_key: str = None, compress: bool = False) -> Tuple[bytes, str]:
    """Convert object to binary"""
    bin_val = pickle.dumps(obj)
    # compression should be done before encrypting, otherwise proper compression cannot be done
    # and size of bytes object might even increase
    if compress:
        bin_val = _compress_bytes(bin_val)
    return EncryptionHelper.encrypt_bytes(bin_val, secret_key=secret_key)


def get_encrypted_binary_str(obj, secret_key: str):
    """Converts object to bytes and then base64 string"""
    objbin, _ = EncryptionHelper.encrypt_bytes(obj, secret_key)
    bytestr = codecs.encode(objbin, "base64").decode("utf-8")
    return bytestr


def get_object_from_encrypted_binary(binary: bytes, secret_key: str, is_compressed: bool = False):
    """Converts bytes to object"""
    bin_val = EncryptionHelper.decrypt_bytes(binary, secret_key)
    if is_compressed:
        try:
            bin_val = _decompress_bytes(bin_val)
        except zlib.error as exc:
            raise DecompressionError from exc
    return pickle.loads(bin_val)  # nosec


def get_object_from_encrypted_binary_str(string: str, secret_key: str):
    """Converts base64 encoded bytes to object"""
    encoded = string.encode("utf-8")
    decoded = codecs.decode(encoded, "base64")
    unpickled = pickle.loads(EncryptionHelper.decrypt_bytes(decoded, secret_key))  # nosec
    return unpickled


def get_attribute_dict(obj, with_class_attrs: bool = False):
    """Returns a dictionary where keys are object attributes and values are attribute values"""

    if not with_class_attrs:
        return obj.__dict__

    members = inspect.getmembers(obj)
    attr_dict = {}
    for member in members:
        name = member[0]
        value = member[1]
        if name.startswith("__") and name.endswith("__"):
            continue
        if not inspect.isroutine(value):
            attr_dict[name] = value
    return attr_dict


def is_null(obj: object) -> bool:
    """
    Checks if the object is null.
    :param obj: object
    :return: bool type to check if the object is null
    """
    return obj is None


def _compress_bytes(value: bytes) -> bytes:
    """Compresses bytes instance"""
    return zlib.compress(value)


def _decompress_bytes(value: bytes) -> bytes:
    """Decompresses bytes instance"""
    return zlib.decompress(value)



"""Includes FeatureSimilaritySelectionOutput class."""
from organon.ml.feature_selection.domain.objects.base_selection_output import BaseSelectionOutput


class FeatureSimilaritySelectionOutput(BaseSelectionOutput):
    """Output of Feature Similarity selection"""



# pylint: skip-file
"""This module keeps the linear algebra enumerations."""
from enum import Enum
import ctypes
import numpy as np


class SparseOperationT(Enum):
    """Enum for sparse matrix operations."""
    SPARSE_OPERATION_NON_TRANSPOSE = 10
    SPARSE_OPERATION_TRANSPOSE = 11
    SPARSE_OPERATION_CONJUGATE_TRANSPOSE = 12


class SparseIndexBaseT(Enum):
    """Enum for sparse matrix index base."""
    SPARSE_INDEX_BASE_ZERO = 0
    SPARSE_INDEX_BASE_ONE = 1


class SparseStatusT(Enum):
    """Enum for sparse matrix operations status."""
    SPARSE_STATUS_SUCCESS = 0
    SPARSE_STATUS_NOT_INITIALIZED = 1
    SPARSE_STATUS_ALLOC_FAILED = 2
    SPARSE_STATUS_INVALID_VALUE = 3
    SPARSE_STATUS_EXECUTION_FAILED = 4
    SPARSE_STATUS_INTERNAL_ERROR = 5
    SPARSE_STATUS_NOT_SUPPORTED = 6


class SparseLayoutT(Enum):
    """Enum for sparse matrix layout type."""
    SPARSE_LAYOUT_ROW_MAJOR = 101
    SPARSE_LAYOUT_COLUMN_MAJOR = 102


class SparseMatrix(ctypes.Structure):
    """Enum for sparse matrix."""


class MatrixDescr(ctypes.Structure):
    """Enum for matrix description."""
    _fields_ = [('type', ctypes.c_int),
                ('mode', ctypes.c_int),
                ('diag', ctypes.c_int)]


descr = MatrixDescr()
descr.type = ctypes.c_int(20)
descr.mode = ctypes.c_int(42)
descr.diag = ctypes.c_int(50)

sparse_matrix_t = ctypes.POINTER(SparseMatrix)

matdescrA = np.chararray((4,))
matdescrA[0] = 'G'
matdescrA[1] = 'L'
matdescrA[2] = 'N'
matdescrA[3] = 'C'

transA = ctypes.c_char_p(b'n')



"""
this module includes encryption helper functions.
"""
import base64
import os
import string
from random import SystemRandom
from typing import Tuple

from cryptography.hazmat.primitives.ciphers.aead import AESGCM

from organon.fl.core.exceptionhandling.known_exception import KnownException


class EncryptionHelper:
    """Holds static methods for encryption/decryption"""

    GCM_IV_LENGTH = 12
    SECRET_KEY_MAX_LENGTH = 16

    @staticmethod
    def generate_key():
        """Generates random key of length 16"""
        all_chars = string.ascii_letters + string.digits + string.punctuation
        return "".join(SystemRandom().sample(all_chars, EncryptionHelper.SECRET_KEY_MAX_LENGTH))

    @staticmethod
    def encrypt_string(value: str, secret_key: str = None) -> str:
        """
        Returns encrypted version of given string
        :param value: string to be encrypted
        :param secret_key: secret key to be used in encryption
        :type value: str
        :return: Encrypted version of string on base64
        """
        byte_val = bytes(value, encoding="utf-8")
        b64_str = base64.b64encode(EncryptionHelper.encrypt_bytes(byte_val, secret_key=secret_key)[0]).decode()
        return b64_str

    @staticmethod
    def decrypt_string(encrypted_str: str, secret_key: str) -> str:
        """
        Decrypts given string
        :param encrypted_str: string to be decrypted
        :param secret_key: secret key to be used in encryption
        :type encrypted_str: str
        :return: Decrypted version of string
        """
        data_bin = base64.b64decode(encrypted_str)
        output = EncryptionHelper.decrypt_bytes(data_bin, secret_key)
        return output.decode(encoding="utf-8")

    @staticmethod
    def __get_normalized_key(key: str) -> bytes:
        max_len = EncryptionHelper.SECRET_KEY_MAX_LENGTH
        lst = bytearray(key, "utf-8")
        key_len = len(lst)
        if key_len > max_len:
            raise KnownException(f"Secret key max length is: {max_len}")
        out = lst.copy()
        out[key_len:] = [0 for _ in range(max_len - key_len)]
        return bytes(out)

    @staticmethod
    def encrypt_bytes(byte_val: bytes, secret_key: str = None) -> Tuple[bytes, str]:
        """
        Encrypts given bytes value.
        :param byte_val: bytes to be encrypted
        :param secret_key: secret key to be used in encryption
        :return: encrypted version of byte_val
        """
        if secret_key is None:
            secret_key = EncryptionHelper.generate_key()
        init_vector = os.urandom(EncryptionHelper.GCM_IV_LENGTH)
        key = EncryptionHelper.__get_normalized_key(secret_key)

        aesgcm = AESGCM(key)
        encrypted = aesgcm.encrypt(init_vector, byte_val, None)

        return init_vector + encrypted, secret_key

    @staticmethod
    def decrypt_bytes(encrypted_bytes: bytes, secret_key: str) -> bytes:
        """
        Decrypts given bytes value
        :param encrypted_bytes: bytes to be decrypted
        :param secret_key: secret key used in encryption
        :return: Decrypted version of encrypted_bytes
        """
        if secret_key is None:
            raise KnownException("Cannot decrypt. Secret_key is None.")
        iv_bytes = encrypted_bytes[:EncryptionHelper.GCM_IV_LENGTH]
        encrypted_data = encrypted_bytes[EncryptionHelper.GCM_IV_LENGTH:]
        key = EncryptionHelper.__get_normalized_key(secret_key)
        aesgcm = AESGCM(key)
        return aesgcm.decrypt(iv_bytes, encrypted_data, None)



"""
This module includes enum class AfeLearningType.
"""
from enum import Enum


# pylint: disable=invalid-name
class AfeLearningType(Enum):
    """
    Learning type for Automated Feature Extraction
    """
    Supervised = 0
    Unsupervised = 1



"""Includes DqFileInputSourceSettings class."""
from typing import List

from organon.idq.domain.businessobjects.record_sources.dq_file_record_source import DqFileRecordSource
from organon.idq.domain.settings.abstractions.dq_base_input_source_settings import DqBaseInputSourceSettings


class DqFileInputSourceSettings(DqBaseInputSourceSettings[DqFileRecordSource]):
    """DQ input source settings for file source type"""

    def __init__(self, source: DqFileRecordSource):
        super().__init__(source)
        self.number_of_rows_per_step: int = None
        self.csv_separator: str = None
        self.date_columns: List[str] = None



"""Includes DqDfDataColumn class."""
import numpy as np

from organon.idq.domain.businessobjects.data_column.dq_data_column import DqDataColumn


class DqDfDataColumn(DqDataColumn):
    """DqDataColumn corresponding to a column in a pandas dataframe"""

    def __init__(self):
        super().__init__()
        self.col_np_dtype: np.dtype = None



"""
This module keeps the ProcessInfoDto class.
"""


class ProcessInfoDto:
    """
    This class keeps the information of process.
    """
    def __init__(self, process_name: str, process_id: int):
        self.process_name = process_name
        self.process_id = process_id



"""Includes AfeFeature class"""
from organon.afe.domain.modelling.businessobjects.afe_column import AfeColumn
from organon.afe.domain.modelling.businessobjects.base_afe_feature import BaseAfeFeature


class AfeFeature(BaseAfeFeature[AfeColumn]):
    """Afe feature information class"""



""" This module includes HighCorrelatedFeatureReduction"""
import random
from typing import List, Tuple, Optional, Dict

import pandas as pd

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.fl.core.helpers.data_frame_helper import get_column_native_type
from organon.ml.feature_reduction.domain.enums.feature_reduction_types import FeatureReductionType
from organon.ml.feature_reduction.domain.helpers.univariate_performance_helper import \
    get_univariate_performances_for_columns
from organon.ml.feature_reduction.domain.objects.high_corr_reduction_output import HighCorrReductionOutput
from organon.ml.feature_reduction.domain.reductions.base_feature_reduction import BaseFeatureReduction
from organon.ml.feature_reduction.settings.objects.high_correlated_feature_reduction_settings import \
    HighCorrelatedFeatureReductionSettings


class HighCorrelatedFeatureReduction(BaseFeatureReduction):
    """HighCorrelatedFeatureReduction class"""

    def _execute_reduction(self, settings: HighCorrelatedFeatureReductionSettings) -> HighCorrReductionOutput:
        data = settings.data
        output = HighCorrReductionOutput()
        output.feature_reduction_type = FeatureReductionType.HIGH_CORRELATION

        included_columns = self._get_included_columns(data, settings.excluded_columns)

        numeric_columns = [col for col in included_columns if
                           (get_column_native_type(data, col) == ColumnNativeType.Numeric)
                           and (col != settings.target_column_name)]
        if not numeric_columns:
            output.reduced_column_list = None
            return output
        high_correlated_column_pair = self._get_high_correlated_columns(data, numeric_columns,
                                                                        settings.correlation_threshold)
        reduced_columns, new_col_performances = self._find_reduced_columns(settings, high_correlated_column_pair)
        output.reduced_column_list = reduced_columns
        output.new_univariate_performance_results = new_col_performances
        return output

    @staticmethod
    def _get_high_correlated_columns(data: pd.DataFrame, numeric_columns: List[str],
                                     correlation_threshold: float) -> List[Tuple[str, str]]:
        """
        Find high correlated numeric features
        Parameters
        ----------
        data
        numeric_columns
        correlation_threshold

        Returns
        -------

        """
        column_pairs = []

        corr_matrix = data[numeric_columns].corr(method="spearman").abs()
        corr_matrix_sorted = corr_matrix[corr_matrix > correlation_threshold].unstack().sort_values(
            ascending=False).dropna()

        for col_x, col_y in corr_matrix_sorted.keys():
            if col_x != col_y and col_x < col_y:
                column_pairs.append((col_x, col_y))

        return column_pairs

    def _find_reduced_columns(self, settings: HighCorrelatedFeatureReductionSettings,
                              high_correlated_column_pair: List[Tuple[str, str]]) \
            -> Tuple[List[str], Optional[Dict[str, float]]]:
        """
        Find reduced columns.
        Parameters
        ----------
        settings
        high_correlated_column_pair


        Returns
        -------
        Returns columns list which reduced.
        """
        if len(high_correlated_column_pair) == 0:
            return [], None
        reduced_columns = []

        if settings.target_column_name is not None and settings.target_type is not None:
            reduced_columns, new_performances = self._get_reduced_columns_via_performance(settings,
                                                                                          high_correlated_column_pair)
            return reduced_columns, new_performances
        to_protect = []
        for column_pair in high_correlated_column_pair:
            value = random.sample(list([column_pair[0], column_pair[1]]), 2)
            column_to_drop = value[0]
            if (column_to_drop not in reduced_columns) and (column_to_drop not in to_protect) and (
                    value[1] not in reduced_columns):
                reduced_columns.append(column_to_drop)
                if value[1] not in to_protect:
                    to_protect.append(value[1])

        return reduced_columns, None

    @staticmethod
    def _get_reduced_columns_via_performance(
            settings: HighCorrelatedFeatureReductionSettings,
            high_correlated_columns_info: List[Tuple[str, str]]) -> Tuple[List[str], Optional[Dict[str, float]]]:
        """
        Finds the reduced column using the univariate performance result.
        Parameters
        ----------
        settings
        high_correlated_columns_info

        Returns
        -------
        Returns columns list which reduced.

        """

        data = settings.data
        target_type = settings.target_type
        target_column_name = settings.target_column_name
        existing_performances = {} if settings.univariate_performance_result is None \
            else settings.univariate_performance_result
        new_performances = {}

        to_protect = []
        reduced_columns = []
        for column_pair in high_correlated_columns_info:
            if column_pair[0] in to_protect and column_pair[1] in to_protect:
                continue
            score_result = {}
            for column in column_pair:
                if column not in existing_performances:
                    column_perf = get_univariate_performances_for_columns(
                        data, [column], target_column_name, target_type, settings.performance_metric,
                        settings.random_state)[column]
                    new_performances[column] = column_perf
                    score_result[column] = column_perf
                else:
                    score_result[column] = existing_performances[column]
            best_score_column = max(score_result, key=score_result.get)
            for column in column_pair:
                if column != best_score_column:
                    if (column not in reduced_columns) and (column not in to_protect):
                        reduced_columns.append(column)
                elif column not in to_protect:
                    to_protect.append(column)

        return reduced_columns, new_performances

    def get_description(self) -> str:
        return "High Correlated Feature Reduction"



"""Includes helper functions for switching dev_mode on/off"""
from organon.fl.core.helpers import guid_helper
from organon.fl.logging.helpers.log_helper import LogHelper, Levels


def init_dev_mode(module_name: str, log_to_console: bool = True, log_file: str = "application.log"):
    """Initializes development mode."""
    execution_id = guid_helper.new_guid(32)

    LogHelper.set_logger_level(Levels.INFO, module_name)
    if log_to_console:
        LogHelper.add_console_handler(module_name)
    if log_file:
        LogHelper.add_file_handler(log_file, module_name)
    LogHelper.add_global_extra("execution_id", execution_id)
    default_format = "[%(asctime)s] [%(threadName)s] %(levelname)-8s %(module_name)s:%(line_number)s : " \
                     "%(execution_id)s  - %(message)s"
    LogHelper.set_default_formatter_default_format_str(default_format)



"""Class for general text classification settings"""
from organon.ml.text_classification.domain.enums.classification_languages import ClassificationLanguages
from organon.ml.text_classification.domain.enums.model_checkpoints import ModelCheckpoints
from organon.ml.text_classification.domain.enums.model_run_type import ModelRunType
from organon.ml.text_classification.domain.objects.grid_search_settings import GridSearchSettings
from organon.ml.text_classification.domain.objects.model_parameter_settings import ModelParameterSettings
from organon.ml.text_classification.domain.objects.optimizer_settings import OptimizerSettings
from organon.ml.text_classification.domain.objects.tokenizer_settings import TokenizerSettings


class TextClassificationSettings:
    """Text Classification Module Settings"""

    def __init__(self, language: ClassificationLanguages, text_column: str = None, target_column: str = None,
                 checkpoint: ModelCheckpoints = ModelCheckpoints.BERT_BASE,
                 mdl_run_type: ModelRunType = ModelRunType.EFFICIENT,
                 mdl_param_settings: ModelParameterSettings = None,
                 opt_settings: OptimizerSettings = None,
                 tokenizer_settings: TokenizerSettings = None,
                 grid_search_settings: GridSearchSettings = None):
        # pylint: disable=too-many-arguments
        self.language = language
        self.text_column = text_column
        self.target_column = target_column
        self.checkpoint: ModelCheckpoints = checkpoint
        self.mdl_run_type = mdl_run_type
        self.mdl_param_settings: ModelParameterSettings = mdl_param_settings \
            if mdl_param_settings is not None else ModelParameterSettings()
        self.opt_settings: OptimizerSettings = opt_settings \
            if opt_settings is not None else OptimizerSettings()
        self.tokenizer_settings: TokenizerSettings = tokenizer_settings \
            if tokenizer_settings is not None else TokenizerSettings()
        self.grid_src_settings: GridSearchSettings = grid_search_settings \
            if grid_search_settings is not None else GridSearchSettings()



"""Includes ObjectClassifier class."""
from typing import Optional, Tuple, Union, Dict, TYPE_CHECKING
from pandas import DataFrame

from organon.ml.common.enums.color_type import ColorType
from organon.ml.common.enums.optimizer_type import OptimizerType
from organon.ml.common.enums.pretrained_model_type import PretrainedModelType
from organon.ml.common.enums.random_flip_type import RandomFlipType
from organon.ml.common.helpers.user_input_service_helper import get_enum
from organon.ml.object_classification.domain.objects.data_augmentation_settings import DataAugmentationSettings
from organon.ml.object_classification.domain.objects.fine_tuning_settings import FineTuningSettings
from organon.ml.object_classification.domain.objects.object_clf_settings import ObjectClfSettings
from organon.ml.object_classification.domain.objects.pretrained_model_settings import PretrainedModelSettings
from organon.ml.object_classification.domain.objects.transfer_learning_settings import TransferLearningSettings
from organon.ml.object_classification.domain.services.binary_object_classificion_service import \
    BinaryObjectClassificationService
from organon.ml.object_classification.domain.services.multiclass_object_classification_service import \
    MulticlassObjectClassificationService
from organon.ml.object_classification.domain.services.object_classification_service_factory import \
    ObjectClassificationServiceFactory

if TYPE_CHECKING:
    import tensorflow as tf


class ObjectClassifier:
    """Image object classifier"""

    def __init__(self, train_data_dir: str, validation_data_dir: str = None,
                 validation_data_ratio: float = 0.2, image_size: Tuple[int, int] = (150, 150),
                 batch_size: int = 50, color_mode: str = ColorType.RGB.name, random_seed: int = 42):
        # pylint: disable=too-many-arguments
        color_mode = get_enum(color_mode, ColorType)
        self._settings = ObjectClfSettings(train_data_dir, validation_data_dir=validation_data_dir,
                                           validation_data_ratio=validation_data_ratio, image_size=image_size,
                                           batch_size=batch_size, color_mode=color_mode, random_seed=random_seed)

        self._service: Union[BinaryObjectClassificationService, MulticlassObjectClassificationService] = None
        self._fine_tuning_settings: FineTuningSettings = None
        self._transfer_lrn_settings: TransferLearningSettings = None
        self._pretrained_model_settings: PretrainedModelSettings = None
        self._data_augmentation_settings: DataAugmentationSettings = None

    def fit(self):
        """Fits classifier"""
        self._set_full_settings()
        self._service = ObjectClassificationServiceFactory.get_classification_service(self._settings)
        self._service.fit()

    def predict(self, data: Union[str, "tf.data.Dataset"], is_processed: bool = False) -> DataFrame:
        """Predict object in image"""
        return self._service.predict(data, is_processed)

    def predict_proba(self, data: Union[str, "tf.data.Dataset"], is_processed: bool = False) -> DataFrame:
        """Predict probabilities for object in image"""
        return self._service.predict_proba(data, is_processed)

    def evaluate(self, data: Union["tf.data.Dataset", str], average: str = None, return_df: bool = True,
                 is_processed: bool = False) -> Tuple[
        Dict[str, float], Optional[DataFrame]]:
        """Predict and evaluate predictions"""
        return self._service.evaluate(data, is_processed=is_processed, average=average, return_df=return_df)

    def get_processed_data(self, data: Union[str, "tf.data.Dataset"]) -> "tf.data.Dataset":
        """Convert data into a format the model can predict"""
        return self._service.get_processed_data(data)

    def set_fine_tuning_settings(self, epoch: int = 10, early_stopping_patience: int = 2,
                                 early_stopping_min_delta=0.001, learning_rate: float = 1e-5,
                                 optimizer: str = OptimizerType.ADAM.name):
        """Sets fine tuning settings for fitting"""
        opt = get_enum(optimizer, OptimizerType)
        self._fine_tuning_settings = FineTuningSettings(epoch, early_stopping_patience, early_stopping_min_delta,
                                                        learning_rate, opt)

    def set_transfer_learning_settings(self, epoch: int = 20, early_stopping: int = 2,
                                       optimizer: str = OptimizerType.ADAM.name,
                                       dropout: float = 0.2, learning_rate: float = 0.001):
        """Sets transfer learning settings for fitting"""
        opt = get_enum(optimizer, OptimizerType)
        self._transfer_lrn_settings = TransferLearningSettings(epoch, early_stopping, opt, dropout, learning_rate)

    def set_pretrained_model_settings(self, model: str = PretrainedModelType.XCEPTION.name,
                                      weights: Optional[str] = "imagenet"):
        """Sets pretrained model settings for fitting"""
        model_enum_val = get_enum(model, PretrainedModelType)
        self._pretrained_model_settings = PretrainedModelSettings(model_enum_val, weights)

    def set_data_augmentation_settings(self, random_flip: str = RandomFlipType.HORIZONTAL_AND_VERTICAL.name,
                                       random_rotation: float = 0.1, random_zoom: float = 0.1):
        """Sets pretrained model settings for fitting"""
        random_flip_type = get_enum(random_flip, RandomFlipType)
        self._data_augmentation_settings = DataAugmentationSettings(random_flip_type, random_rotation, random_zoom)

    def _set_full_settings(self):
        if self._fine_tuning_settings is not None:
            self._settings.fine_tuning_settings = self._fine_tuning_settings
        if self._transfer_lrn_settings is not None:
            self._settings.transfer_learning_settings = self._transfer_lrn_settings
        if self._pretrained_model_settings is not None:
            self._settings.pretrained_model_settings = self._pretrained_model_settings
        if self._data_augmentation_settings is not None:
            self._settings.data_augmentation_settings = self._data_augmentation_settings



"""Includes BinaryReporterService class."""
from typing import Tuple

import numpy as np
import pandas as pd
from sklearn.metrics import roc_auc_score

from organon.ml.common.helpers.df_ops_helper import get_bins
from organon.ml.reporting.domain.objects.binary_report import BinaryReport
from organon.ml.reporting.domain.services.base_reporter_service import BaseReporterService
from organon.ml.reporting.settings.objects.reporter_settings import ReporterSettings


class BinaryReporterService(BaseReporterService[BinaryReport]):
    """Reporter service for binary classification"""

    @classmethod
    def execute(cls, settings: ReporterSettings) -> BinaryReport:
        group_by_columns = []
        id_str_col_exists = settings.id_str_column is not None
        split_col_exists = settings.split_column is not None

        if id_str_col_exists:
            group_by_columns.append(settings.id_str_column)
        if split_col_exists:
            group_by_columns.append(settings.split_column)

        perf_sum_columns = cls.__get_columns_for_performance_summary_df(id_str_col_exists, split_col_exists)
        detailed_perf_sum_columns = cls.__get_columns_for_detailed_performance_summary_df(id_str_col_exists,
                                                                                          split_col_exists)
        target_based_summary_columns = cls.__get_columns_for_target_based_performance_summary_df(id_str_col_exists,
                                                                                                 split_col_exists)
        if group_by_columns:
            performance_summary_dicts = []
            all_bin_performances = []
            all_target_based_performances = []
            for groups, indices in settings.data.groupby(group_by_columns).indices.items():
                group_tuple = groups if isinstance(groups, tuple) else (groups,)
                general_performance_dict_for_group, bin_performances, target_based_score_perf = \
                    cls.__get_details(settings, settings.data.loc[indices])
                cls.__add_group_values_to_dict(general_performance_dict_for_group, group_tuple, id_str_col_exists,
                                               split_col_exists)
                cls.__add_group_values_to_dict(target_based_score_perf, group_tuple, id_str_col_exists,
                                               split_col_exists)
                for _dict in bin_performances:
                    cls.__add_group_values_to_dict(_dict, group_tuple, id_str_col_exists,
                                                   split_col_exists)

                performance_summary_dicts.append(general_performance_dict_for_group)
                all_bin_performances.extend(bin_performances)
                all_target_based_performances.append(target_based_score_perf)
            performance_summary_df = pd.DataFrame(performance_summary_dicts, columns=perf_sum_columns)
            detailed_performance_summary_df = pd.DataFrame(all_bin_performances,
                                                           columns=detailed_perf_sum_columns)
            target_based_summary_df = pd.DataFrame(all_target_based_performances,
                                                   columns=target_based_summary_columns)
        else:
            general_performance, bin_performances, target_based_score_perf = cls.__get_details(settings, settings.data)
            performance_summary_df = pd.DataFrame([general_performance], columns=perf_sum_columns)
            detailed_performance_summary_df = pd.DataFrame(bin_performances, columns=detailed_perf_sum_columns)
            target_based_summary_df = pd.DataFrame([target_based_score_perf],
                                                   columns=target_based_summary_columns)

        report = BinaryReport()
        report.performance_summary = performance_summary_df.T
        report.detailed_performance_summary = detailed_performance_summary_df
        report.target_based_performance_summary = target_based_summary_df.T
        return report

    @classmethod
    def __get_details(cls, settings, all_data):
        total_row_count = len(all_data)
        if settings.num_bins > total_row_count:
            raise ValueError("num_bins cannot be higher than number of rows in a set")
        bins = get_bins(all_data[settings.score_column], settings.num_bins)

        bin_dicts = []
        total_neg_count = 0
        total_pos_count = 0
        score_col = settings.score_column

        for bin_num in range(settings.num_bins):
            data = all_data.loc[bins == bin_num]
            total_count, positive_count, negative_count = cls.__get_row_positive_negative_count(data,
                                                                                                settings.target_column)

            bin_dicts.append({
                "BIN": bin_num,
                "TOTAL COUNT": total_count,
                "POSITIVE COUNT": positive_count,
                "NEGATIVE COUNT": negative_count,
                "TARGET RATIO": positive_count / total_count,
                "SCORE AVG": data[score_col].mean(),
                "SCORE SUM": data[score_col].sum(),
                "SCORE MIN": data[score_col].min(),
                "SCORE MAX": data[score_col].max(),
                "TRUE NEGATIVE COUNT": total_neg_count,
                "FALSE NEGATIVE COUNT": total_pos_count,
            })
            total_pos_count += positive_count
            total_neg_count += negative_count

        true_positive_count = 0
        false_positive_count = 0

        general_performance_dict = cls._get_general_performance_dict(all_data, settings, total_row_count,
                                                                     total_pos_count, total_neg_count)
        total_pos_ratio = general_performance_dict["POSITIVE RATIO"]
        pos_count_sum = 0
        total_count_sum = 0
        for bin_num in range(settings.num_bins - 1, -1, -1):
            bin_dict = bin_dicts[bin_num]
            true_positive_count += bin_dict["POSITIVE COUNT"]
            false_positive_count += bin_dict["NEGATIVE COUNT"]
            bin_dict["TRUE POSITIVE COUNT"] = true_positive_count
            bin_dict["FALSE POSITIVE COUNT"] = false_positive_count
            bin_dict["TRUE POSITIVE RATE"] = bin_dict["TRUE POSITIVE COUNT"] / (
                    bin_dict["TRUE POSITIVE COUNT"] + bin_dict["FALSE NEGATIVE COUNT"])
            bin_dict["FALSE POSITIVE RATE"] = bin_dict["FALSE POSITIVE COUNT"] / (
                    bin_dict["FALSE POSITIVE COUNT"] + bin_dict["TRUE NEGATIVE COUNT"])
            bin_dict["PRECISION"] = bin_dict["TRUE POSITIVE COUNT"] / (
                    bin_dict["TRUE POSITIVE COUNT"] + bin_dict["FALSE POSITIVE COUNT"])
            bin_dict["ACCURACY"] = (bin_dict["TRUE POSITIVE COUNT"] + bin_dict["TRUE NEGATIVE COUNT"]) / \
                                   ((bin_dict["TRUE POSITIVE COUNT"] + bin_dict["TRUE NEGATIVE COUNT"]) +
                                    (bin_dict["FALSE POSITIVE COUNT"] + bin_dict["FALSE NEGATIVE COUNT"]))
            bin_dict["F1 SCORE"] = cls.__get_f1_score(bin_dict["PRECISION"], bin_dict["TRUE POSITIVE RATE"])
            pos_count_sum += bin_dict["POSITIVE COUNT"]
            total_count_sum += bin_dict["TOTAL COUNT"]
            bin_dict["CUMULATIVE LIFT"] = pos_count_sum / total_count_sum / total_pos_ratio
        target_based_score_perf = cls._get_target_based_score_performance(settings, all_data, general_performance_dict)
        return general_performance_dict, bin_dicts, target_based_score_perf

    @classmethod
    def __get_f1_score(cls, precision, recall):
        if precision + recall != 0:
            return 2 * precision * recall / (precision + recall)
        return 0

    @classmethod
    def _get_target_based_score_performance(cls, settings, all_data: pd.DataFrame, general_performance_dict: dict):
        pos_count = general_performance_dict["POSITIVE COUNT"]
        neg_count = general_performance_dict["NEGATIVE COUNT"]
        pos_dict = cls._get_target_based_score_performance_for_target(settings, all_data, 1, "POSITIVES", pos_count)
        neg_dict = cls._get_target_based_score_performance_for_target(settings, all_data, 0, "NEGATIVES", neg_count)
        pos_dict.update(neg_dict)
        return pos_dict

    @classmethod
    def _get_target_based_score_performance_for_target(cls, settings, all_data: pd.DataFrame, target_val: int,
                                                       target_key_str: str,
                                                       count: int):
        score_col = settings.score_column
        pos_indices = all_data[settings.target_column] == target_val
        pos_data: pd.DataFrame = all_data[score_col].loc[pos_indices]

        percentile_nums = [0.05 * i for i in range(1, 20)]
        pos_percentiles = pos_data.quantile(percentile_nums)
        perf_dict = {
            f"{target_key_str} COUNT": count,
            f"{target_key_str} STD": pos_data.std(),
            f"{target_key_str} MEAN": pos_data.mean(),
            f"{target_key_str} MIN": pos_data.min(),
            f"{target_key_str} MAX": pos_data.max(),
        }

        for perc_num in percentile_nums:
            perf_dict[f"{target_key_str} P{int(perc_num * 100)}"] = pos_percentiles.loc[perc_num]
        return perf_dict

    @classmethod
    def _get_general_performance_dict(cls, all_data, settings,
                                      total_row_count: int, total_pos_count: int, total_neg_count: int):
        score_col = settings.score_column
        aggregates = all_data[score_col].agg(["mean", "std", "min", "max", "median"])
        percentiles = np.percentile(all_data[score_col], [25, 50, 75])
        general_performance_dict = {
            "ROW COUNT": total_row_count,
            "POSITIVE COUNT": total_pos_count,
            "NEGATIVE COUNT": total_neg_count,
            "POSITIVE RATIO": total_pos_count / total_row_count,
            "NEGATIVE RATIO": total_neg_count / total_row_count,
            "ODDS": total_pos_count / total_neg_count,
            "AUC": roc_auc_score(all_data[settings.target_column], all_data[score_col]),
            "SCORE AVG": aggregates["mean"],
            "SCORE STD": aggregates["std"],
            "SCORE MIN": aggregates["min"],
            "SCORE MAX": aggregates["max"],
            "SCORE P25": percentiles[0],
            "SCORE MEDIAN": percentiles[1],
            "SCORE P75": percentiles[2],
        }
        return general_performance_dict

    @classmethod
    def __add_group_values_to_dict(cls, dict_to_add: dict, group_tuple, id_str_col_exists: bool,
                                   split_col_exists: bool):
        if id_str_col_exists:
            dict_to_add["ID_STR"] = group_tuple[0]
            if split_col_exists:
                dict_to_add["SPLIT"] = group_tuple[1]
        elif split_col_exists:
            dict_to_add["SPLIT"] = group_tuple[0]

    @classmethod
    def __get_columns_for_performance_summary_df(cls, id_str_col_exists: bool, split_col_exists: bool):
        columns = ["AUC", "ROW COUNT", "POSITIVE COUNT", "NEGATIVE COUNT", "POSITIVE RATIO",
                   "NEGATIVE RATIO", "ODDS", "SCORE AVG", "SCORE STD", "SCORE MIN", "SCORE P25",
                   "SCORE MEDIAN", "SCORE P75", "SCORE MAX"]
        if split_col_exists:
            columns = ["SPLIT"] + columns
        if id_str_col_exists:
            columns = ["ID_STR"] + columns
        return columns

    @classmethod
    def __get_columns_for_detailed_performance_summary_df(cls, id_str_col_exists: bool, split_col_exists: bool):
        columns = ["BIN", "POSITIVE COUNT", "NEGATIVE COUNT", "TOTAL COUNT",
                   "TARGET RATIO", "SCORE AVG", "TRUE POSITIVE COUNT", "FALSE POSITIVE COUNT", "TRUE NEGATIVE COUNT",
                   "FALSE NEGATIVE COUNT", "TRUE POSITIVE RATE", "FALSE POSITIVE RATE", "PRECISION", "ACCURACY",
                   "F1 SCORE", "CUMULATIVE LIFT"]
        if split_col_exists:
            columns = ["SPLIT"] + columns
        if id_str_col_exists:
            columns = ["ID_STR"] + columns
        return columns

    @classmethod
    def __get_columns_for_target_based_performance_summary_df(cls, id_str_col_exists: bool, split_col_exists: bool):
        suffixes = ["COUNT", "STD", "MEAN", "MIN"]
        suffixes.extend([f"P{5 * i}" for i in range(1, 20)])
        suffixes.append("MAX")
        columns = [f"{prefix} {suffix}" for prefix in ["POSITIVES", "NEGATIVES"] for suffix in suffixes]
        if split_col_exists:
            columns = ["SPLIT"] + columns
        if id_str_col_exists:
            columns = ["ID_STR"] + columns
        return columns

    @classmethod
    def __get_row_positive_negative_count(cls, data: pd.DataFrame, target_col: str) -> Tuple[int, int, int]:
        row_count = cls._get_row_count(data)
        positive_count = np.sum(data[target_col] == 1)
        negative_count = row_count - positive_count
        return row_count, positive_count, negative_count



"""This module includes ScoringService class."""
from datetime import datetime
from typing import List, Dict, TypeVar, Generic

import numpy as np
import pandas as pd

from organon.afe.core.businessobjects.afe_static_objects import AfeStaticObjects
from organon.afe.dataaccess.services.transaction_file_record_reader import TransactionFileRecordReader
from organon.afe.domain.common import afe_date_helper
from organon.afe.domain.common.afe_date_helper import get_str_to_date_converter
from organon.afe.domain.enums.afe_date_column_type import AfeDateColumnType
from organon.afe.domain.enums.afe_operator import AfeOperator
from organon.afe.domain.enums.binary_target_class import BinaryTargetClass
from organon.afe.domain.enums.date_resolution import DateResolution
from organon.afe.domain.modelling.businessobjects.data_frame_builder import DataFrameBuilder
from organon.afe.domain.modelling.businessobjects.multi_target_entity_container import MultiTargetEntityContainer
from organon.afe.domain.modelling.businessobjects.target_file_record import TargetFileRecord
from organon.afe.domain.modelling.businessobjects.target_file_record_collection import TargetFileRecordCollection
from organon.afe.domain.modelling.businessobjects.transaction_file_record_collection import \
    TransactionFileRecordCollection
from organon.afe.domain.modelling.businessobjects.transaction_file_stats import TransactionFileStats
from organon.afe.domain.modelling.supervised.multi_target_entity_container_service import \
    MultiTargetEntityContainerService
from organon.afe.domain.settings.base_afe_scoring_settings import BaseAfeScoringSettings
from organon.afe.domain.settings.record_source import RecordSource
from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.fl.core.helpers import list_helper, date_helper
from organon.fl.logging.helpers.log_helper import LogHelper

AfeScoringSettingsType = TypeVar("AfeScoringSettingsType", bound=BaseAfeScoringSettings)


class AfeScoringService(Generic[AfeScoringSettingsType]):
    """Afe Scoring Service class."""

    def __init__(self, scoring_settings: AfeScoringSettingsType):
        self.scoring_settings = scoring_settings

    def score(self, num_of_threads: int = None, for_date: str or int or datetime = None):
        """Score afe using given number of threads

        :param for_date: if given, scoring will be done for given date and no target file will be used
        :param int num_of_threads: Number of threads/processes to run scoring in parallel
        """
        if num_of_threads is None:
            num_of_threads = self.scoring_settings.process_settings.number_of_cores

        return self.score_in_memory(num_of_threads, for_date=for_date)

    def score_in_memory(self, num_of_threads: int, for_date: datetime = None) -> pd.DataFrame:
        """Scores afe in-memory"""

        if for_date is not None and not isinstance(for_date, datetime):
            try:
                target_date_col = self.scoring_settings.model_output.target_descriptor.date_column
                converter = get_str_to_date_converter(target_date_col.date_column_type, target_date_col.custom_format)
                for_date = converter(str(for_date))
            except Exception as exc:
                raise KnownException("for_date was not given in target column format and type") from exc

        LogHelper.info("Scoring in-memory")

        features = list(self.scoring_settings.model_output.output_features.values())

        LogHelper.info("Started creating data structures needed prior to reading transactions file")

        d_columns = self.__update_column_list(list({feature.afe_column.dimension_name for feature in features}),
                                              AfeStaticObjects.empty_dimension_column)
        q_columns = self.__update_column_list(list({feature.afe_column.quantity_name for feature in features}),
                                              AfeStaticObjects.empty_quantity_column)

        trx_stats_from_modelling = self.scoring_settings.model_output.transaction_file_stats

        max_horizon_per_date_col = self._get_max_horizon_per_date_col()

        if for_date is None:
            multi_target_entity_container_service = self._get_multi_target_entity_container_service()
            multi_target_entity_container_service.execute()
            multi_target_entity_container = multi_target_entity_container_service.container
            target_file_record_collection = list(multi_target_entity_container.records_per_file.values())[0]
            trx_file_record_collection = self._get_transaction_file_record_collection(d_columns,
                                                                                      q_columns,
                                                                                      trx_stats_from_modelling,
                                                                                      max_horizon_per_date_col,
                                                                                      target_file_record_collection,
                                                                                      multi_target_entity_container
                                                                                      )
        else:
            trx_file_record_collection = self.__get_transaction_file_record_collection_for_score_date(
                d_columns,
                q_columns,
                trx_stats_from_modelling,
                max_horizon_per_date_col,
                for_date)
            target_file_record_collection = TargetFileRecordCollection(len(trx_file_record_collection.entity_index_map))
            for_date_int = date_helper.get_date_as_integer(for_date)
            for entity in trx_file_record_collection.entity_index_map:
                record = TargetFileRecord()
                record.entity_id = entity
                record.event_date = for_date_int
                record.target_binary = BinaryTargetClass.NAN
                target_file_record_collection.append(record)

        frame_builder = self._get_data_frame_builder(trx_file_record_collection, target_file_record_collection,
                                                     trx_stats_from_modelling)
        frame_builder.execute_for_afe_columns(num_of_threads, self.scoring_settings.model_output.output_features)
        data_frame = frame_builder.get_frame_as_pandas_dataframe()
        target_descriptor = self.scoring_settings.model_output.target_descriptor

        data_frame.rename(
            columns={
                AfeStaticObjects.distinct_entities_entity_column_name: target_descriptor.entity_column_name
            },
            inplace=True)
        if target_descriptor.date_column is not None:
            date_col_name = target_descriptor.date_column.column_name
            data_frame.rename(
                columns={
                    AfeStaticObjects.event_date_column_name: date_col_name
                },
                inplace=True)
            if target_descriptor.date_column.date_column_type != AfeDateColumnType.DateTime:
                converter_func = afe_date_helper.get_date_to_str_converter(
                    target_descriptor.date_column.date_column_type, target_descriptor.date_column.custom_format)
                data_frame[date_col_name] = data_frame[date_col_name].apply(converter_func)

        data_frame.rename(columns={col: col.upper() for col in data_frame.columns}, inplace=True)

        data_frame_new = AfeScoringService.__change_value_for_mode_operator(frame_builder,
                                                                            trx_stats_from_modelling, data_frame)

        return data_frame_new

    def _get_multi_target_entity_container_service(self):
        return MultiTargetEntityContainerService(
            [self.scoring_settings.target_record_source],
            self.scoring_settings.model_output.target_descriptor, is_scoring=True)

    def _get_data_frame_builder(self, trx_file_record_collection: TransactionFileRecordCollection,
                                target_file_record_collection: TargetFileRecordCollection,
                                transaction_file_stats: TransactionFileStats) -> DataFrameBuilder:
        frame_builder = DataFrameBuilder(self.scoring_settings.model_output.target_descriptor,
                                         target_file_record_collection,
                                         trx_file_record_collection,
                                         transaction_file_stats=transaction_file_stats,
                                         max_num_of_columns=len(self.scoring_settings.model_output.output_features))
        return frame_builder

    def _get_max_horizon_per_date_col(self) -> Dict[str, int]:
        max_horizon_per_date_col = {}
        for feature in self.scoring_settings.model_output.output_features.values():
            if feature.afe_column.date_column is not None:
                date_col = feature.afe_column.date_column
                max_time_window = feature.afe_column.time_window
                if date_col.column_name in max_horizon_per_date_col:
                    prev_max = max_horizon_per_date_col[date_col.column_name]
                    max_horizon_per_date_col[date_col.column_name] = max(prev_max, max_time_window)
                else:
                    max_horizon_per_date_col[date_col.column_name] = max_time_window
        return max_horizon_per_date_col

    @staticmethod
    def __change_value_for_mode_operator(frame_builder, trx_file_stats, data_frame):
        for column_name, afe_column in frame_builder.name_to_column.items():
            histogram = trx_file_stats.get_histogram(afe_column.dimension_name)
            if AfeOperator.Mode.name == afe_column.operator.name:
                new_column_name = frame_builder.new_col_names_map[column_name]
                data_frame[new_column_name] = data_frame[new_column_name] \
                    .apply(lambda x, h=histogram: h.reverse_index[int(x)] if not pd.isna(x) else x)
        return data_frame

    def _get_transaction_file_record_collection(self,
                                                d_columns: List[str], q_columns: List[str],
                                                trx_stats_from_modelling: TransactionFileStats,
                                                max_horizon_per_date_col: Dict[str, int],
                                                target_file_record_collection: TargetFileRecordCollection = None,
                                                multi_target_entity_container: MultiTargetEntityContainer = None):

        entity_col_name = AfeStaticObjects.distinct_entities_entity_column_name
        distinct_entity_list = np.unique(target_file_record_collection.
                                         entities[~pd.isnull(target_file_record_collection.entities)]).tolist()
        distinct_entities_df = pd.DataFrame(
            {entity_col_name: distinct_entity_list})
        distinct_entities_record_source = RecordSource(source=distinct_entities_df)
        interval_per_entity_per_date_col = {}

        date_columns = {}

        distinct_date_columns_dict = {
            feature.afe_column.date_column.column_name: (feature.afe_column.date_resolution, feature.afe_column.offset)
            for feature in self.scoring_settings.model_output.output_features.values()
            if feature.afe_column.date_column is not None}

        for date_column_name, (resolution, offset) in distinct_date_columns_dict.items():
            max_horizon = max_horizon_per_date_col[date_column_name]
            interval_per_entity = multi_target_entity_container.get_date_interval_per_entity(resolution,
                                                                                             max_horizon, offset)
            interval_per_entity_per_date_col[date_column_name] = interval_per_entity
            afe_date_col = self.scoring_settings.model_output.trx_date_columns[date_column_name]
            date_columns[date_column_name] = afe_date_col

        empty_col_interval_per_entity = multi_target_entity_container.get_date_interval_per_entity(
            DateResolution.Day, 1, 0)
        interval_per_entity_per_date_col["__NO_DATE_COL"] = empty_col_interval_per_entity

        transaction_file_reader = self._get_transaction_file_record_reader(date_columns,
                                                                           None,
                                                                           interval_per_entity_per_date_col,
                                                                           d_columns, q_columns,
                                                                           trx_stats_from_modelling,
                                                                           distinct_entities_record_source,
                                                                           entity_col_name)

        LogHelper.info("Started reading transactions file.")
        trx_file_record_collection: TransactionFileRecordCollection = transaction_file_reader.read()
        LogHelper.info("Finished reading transactions file.")

        return trx_file_record_collection

    def __get_transaction_file_record_collection_for_score_date(self,
                                                                d_columns: List[str], q_columns: List[str],
                                                                trx_stats_from_modelling: TransactionFileStats,
                                                                max_horizon_per_date_col: Dict[str, int],
                                                                score_date: datetime):
        date_columns = self.scoring_settings.model_output.trx_date_columns
        interval_per_date_col = {}
        for feature in self.scoring_settings.model_output.output_features.values():
            date_col = feature.afe_column.date_column
            if date_col is not None and date_col.column_name not in interval_per_date_col:
                max_horizon = max_horizon_per_date_col[date_col.column_name]
                func_min = afe_date_helper.get_date_subtraction_func(feature.afe_column.date_resolution,
                                                                     max_horizon + feature.afe_column.offset)
                func_max = afe_date_helper.get_date_subtraction_func(feature.afe_column.date_resolution,
                                                                     feature.afe_column.offset)
                date_as_int = date_helper.get_date_as_integer(score_date)
                min_date = date_helper.get_integer_as_date(func_min(date_as_int))
                max_date = date_helper.get_integer_as_date(func_max(date_as_int))
                interval_per_date_col[date_col.column_name] = (min_date, max_date)

        transaction_file_reader = self._get_transaction_file_record_reader(date_columns,
                                                                           interval_per_date_col,
                                                                           None,
                                                                           d_columns, q_columns,
                                                                           trx_stats_from_modelling,
                                                                           None, None)

        LogHelper.info("Started reading transactions file.")
        trx_file_record_collection: TransactionFileRecordCollection = transaction_file_reader.read()
        LogHelper.info("Finished reading transactions file.")
        return trx_file_record_collection

    def _get_transaction_file_record_reader(self, date_columns, interval_per_date_col,
                                            interval_per_entity_per_date_col,
                                            d_columns, q_columns, trx_stats_from_modelling,
                                            distinct_entities_record_source, entity_col_name
                                            ):
        # pylint: disable=too-many-arguments
        return TransactionFileRecordReader(
            self.scoring_settings.raw_input_source,
            date_columns,
            self.scoring_settings.model_output.trx_entity_column,
            interval_per_entity_per_date_column=interval_per_entity_per_date_col,
            interval_per_date_column=interval_per_date_col,
            dimension_columns=d_columns, quantity_columns=q_columns,
            distinct_entity_source=distinct_entities_record_source,
            entity_source_entity_column=entity_col_name,
            transaction_file_stats=trx_stats_from_modelling,
            reading_settings=self.scoring_settings.trx_reading_settings
        )

    @staticmethod
    def __update_column_list(base_list: List[str], new_column: str) -> List[str]:
        if list_helper.is_null_or_empty(base_list):
            return [new_column]
        if new_column not in base_list:
            base_list.append(new_column)
        return list(set(base_list))



"""This module includes helper function for creating dashboards with plotly."""
import webbrowser
from http.server import BaseHTTPRequestHandler, HTTPServer
from typing import List, Tuple

import pandas as pd
import plotly.graph_objects as go
import six
from plotly.basedatatypes import BaseTraceType
from plotly.subplots import make_subplots

ORG_COLOR_HEX = "#2D1A44"


def open_html_in_browser(html, using=None, new=0, autoraise=True):
    """
    Display html in a web browser without creating a temp file.

    Instantiates a trivial http server and uses the webbrowser module to
    open a URL to retrieve html from that server.

    Parameters
    ----------
    html: str
        HTML string to display
    using, new, autoraise:
        See docstrings in webbrowser.get and webbrowser.open
    """
    if isinstance(html, six.string_types):
        html = html.encode("utf8")

    class OneShotRequestHandler(BaseHTTPRequestHandler):
        """Simple http request handler"""
        def do_GET(self):   # pylint: disable=invalid-name
            """Returns html string as response"""
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()

            buffer_size = 1024 * 1024
            for i in range(0, len(html), buffer_size):
                self.wfile.write(html[i: i + buffer_size])

        def log_message(self, _format, *args):  # pylint: disable=arguments-differ
            """Silence stderr logging"""

    server = HTTPServer(("127.0.0.1", 0), OneShotRequestHandler)
    webbrowser.get(using).open(f"http://127.0.0.1:{server.server_port}", new=new, autoraise=autoraise)
    server.handle_request()


def figures_to_html(figs):
    """Merge figure html strings in single html string"""
    html_str = "<html><head></head><body>\n"
    for fig in figs:
        html_str += fig.to_html().split('<body>')[1].split('</body>')[0]
    html_str += "</body></html>\n"
    return html_str


def get_line_scatter_plot(x_vals, y_vals, title: str = None) -> go.Scatter:
    """
    Creates a line plot where data points are marked
    :param y_vals: Y axis values of data points
    :param x_vals: X axis values of data points
    :param title: Title of plot
    :return: Plotly Scatter object with given values
    """
    trace = go.Scatter(x=x_vals, y=y_vals, name=title, mode="markers+lines")
    return trace


def get_data_table_plot(data_frame: pd.DataFrame) -> go.Table:
    """
    Creates plotly table from given data frame
    :param data_frame:
    :return:
    """
    columns = list(data_frame.columns)

    row_even_color = 'lightgrey'
    row_odd_color = 'white'

    table = go.Table(
        header=dict(
            values=[f"<b>{col}</b>" for col in columns],
            line_color='darkslategray',
            fill_color=ORG_COLOR_HEX,
            align=['left', 'center'],
            font=dict(color='white', size=12)
        ),
        cells=dict(
            values=[data_frame[col] for col in columns],
            line_color='darkslategray',
            # 2-D list of colors for alternating rows
            fill_color=[[row_odd_color, row_even_color, row_odd_color, row_even_color,
                         row_odd_color] * len(data_frame.index)],
            align=['left', 'center'],
            font=dict(color='darkslategray', size=10)
        )

    )

    return table


def get_bar_chart(x_vals: List, y_vals: List) -> go.Bar:
    """
    Creates a bar chart with given bar names and values
    :param x_vals: Names of bars
    :param y_vals: Values of bars
    :return: Plotly Bar object with given values
    """
    bar_chart = go.Bar(x=x_vals, y=y_vals)
    return bar_chart


def get_bar_chart_with_width(sorted_bin_limits: List[float], sorted_bin_values: List[float]) \
        -> Tuple[go.Bar, Tuple[float, float], Tuple[float, float]]:
    """Creates bar chart with adjacent bars
    :param sorted_bin_limits: upper limits of bars
    :param sorted_bin_values: values of bars
    :return: Plotly Bar object with given values
    """
    widths = []
    x_vals = []
    y_vals = []
    customdata = []

    for i, limit in enumerate(sorted_bin_limits):
        if i == 0:
            continue
        width = limit - sorted_bin_limits[i - 1]
        widths.append(width)
        y_vals.append(sorted_bin_values[i - 1])
        x_vals.append(limit - width / 2)
        customdata.append(f"({sorted_bin_limits[i - 1]}, {limit})")

    # y_vals = [(-1) ** i * val for i, val in enumerate(y_vals)]

    # calculate y range
    maxy = max(y_vals)
    miny = min(y_vals)
    gap = maxy - miny
    y_min_range = min(0.0, miny - 0.1 * gap)
    y_max_range = max(0.0, maxy + 0.1 * gap)
    y_range = (y_min_range, y_max_range)

    if len(widths) < 3:
        scale_width = min(widths)
    else:
        scale_width = max(widths[1:-1])

    allowed_max_width = scale_width * 100

    x_min_range = sorted_bin_limits[0]
    x_max_range = sorted_bin_limits[-1]

    if widths[0] > allowed_max_width:
        widths[0] = allowed_max_width
        second_bin_start = x_vals[1] - widths[1] / 2
        x_vals[0] = second_bin_start - allowed_max_width / 2
        x_min_range = second_bin_start - scale_width * 2
    if widths[-1] > allowed_max_width:
        widths[-1] = allowed_max_width
        prev_last_bin_end = x_vals[-2] + widths[-2] / 2
        x_vals[-1] = prev_last_bin_end + allowed_max_width / 2
        x_max_range = prev_last_bin_end + scale_width * 2

    extra_space_length = 0.1 * (x_max_range - x_min_range)
    x_min_range -= extra_space_length
    x_max_range += extra_space_length
    x_range = (x_min_range, x_max_range)

    bar_chart = go.Bar(
        x=x_vals,
        y=y_vals,
        customdata=customdata,
        width=widths,
        hovertemplate=
        '%{customdata}' +
        '<extra><b>%{y}</b></extra>'
    )
    return bar_chart, x_range, y_range

    # return show_plot([bar], title, x_axis_title, y_axis_title, x_range, y_range)


# pylint: disable=too-many-arguments
def create_figure(traces: List[BaseTraceType], title=None, x_axis_title=None, y_axis_title=None, x_range=None,
                  y_range=None) -> go.Figure:
    """
    Creates a plotly figure with given traces
    :param traces: plotly objects to be included in figure
    :param title: title of figure
    :param x_axis_title:  x axis label
    :param y_axis_title: y axis label/s
    :param x_range: x axis range to be shown
    :param y_range: y axis range to be shown
    :return:
    """
    if len(traces) == 2:
        fig = make_subplots(specs=[[{"secondary_y": True}]])
        fig.add_trace(
            traces[0],
            secondary_y=False
        )
        fig.add_trace(
            traces[1],
            secondary_y=True
        )
    else:
        fig = go.Figure()
        for trace in traces:
            fig.add_trace(trace)
    if title:
        fig.update_layout(title=title)
    if x_axis_title:
        fig.update_layout(xaxis_title=x_axis_title)
    if y_axis_title:
        if isinstance(y_axis_title, list):
            fig.update_yaxes(title_text=y_axis_title[0], secondary_y=False)
            fig.update_yaxes(title_text=y_axis_title[1], secondary_y=True)
        else:
            fig.update_layout(yaxis_title=y_axis_title)
    if x_range:
        fig.update_layout(xaxis={"range": x_range})
    if y_range:
        fig.update_layout(yaxis={"range": y_range})

    return fig



"""Includes LightGBMRegressor class."""
from lightgbm import LGBMRegressor

from organon.ml.modelling.algorithms.core.abstractions.base_sklearn_regressor import BaseSklearnRegressor


class LightGBMRegressor(BaseSklearnRegressor[LGBMRegressor]):
    """Wrapper for LGBMRegressor"""

    @property
    def regressor_class(self):
        return LGBMRegressor



"""Includes Dq user input classes."""
from typing import List

from organon.idq.services.user_settings.base_dq_user_input import BaseDqUserInput
from organon.idq.services.user_settings.user_multi_partition_calculation_params import UserMultiPartitionCalcParams


class DqUserInput(BaseDqUserInput):
    """User class for all dq settings"""

    def __init__(self):
        super().__init__()
        self.calc_params: UserMultiPartitionCalcParams = None

    def set_multi_partition_calculation(self,
                                        input_source_settings: dict,
                                        max_nominal_cardinality_count: int = None,
                                        outlier_parameter: float = None,
                                        use_population_nominal_stats: bool = None,
                                        name: str = None,
                                        partitions: List[List[dict]] = None):
        """
        Adds calculation settings
        """
        self._check_simple_types(locals(), {
            "input_source_settings": dict,
            "max_nominal_cardinality_count": int,
            "outlier_parameter": (float, int),
            "name": str,
            "partitions": list
        })  # bu satır method'ta en üst satır olmalı (locals kullanımından dolayı)
        self.calc_params = self._get_multi_partition_calculation_params(name, input_source_settings,
                                                                        max_nominal_cardinality_count,
                                                                        outlier_parameter,
                                                                        use_population_nominal_stats,
                                                                        partitions=partitions)



""" This module includes PandasDataFrameOperations."""
import math
from typing import Dict, Any, List

import pandas as pd

from organon.fl.core.businessobjects.dataframe import DataFrame
from organon.fl.mathematics.helpers.idataframe_operations import IDataFrameOperations


class PandasDataFrameOperations(IDataFrameOperations):
    """Class for Pandas Dataframe Operations."""

    @staticmethod
    def join(df1: DataFrame, df2: DataFrame):
        """:returns pandas join function equivalent"""
        joined_df = DataFrame()
        joined_df.data_frame = df1.data_frame.join(df2.data_frame).fillna(0)
        return joined_df

    @staticmethod
    def get_column_stability(df_obj: DataFrame):
        """:returns boolean df depends on whether a column is stable or not"""
        df_to_numpy = df_obj.data_frame.to_numpy()
        return (df_to_numpy[0] == df_to_numpy).all(0) | df_obj.data_frame.isnull().all()

    @staticmethod
    def get_column_distinct_counts(df_obj: DataFrame, col_names: List[str] = None) -> Dict[str, int]:
        """:returns number of unique values per column"""
        df_to_pandas_df = df_obj.data_frame
        cols = col_names if col_names is not None else df_to_pandas_df.columns
        return {col: df_to_pandas_df[col].nunique(dropna=False) for col in cols}

    @staticmethod
    def get_min_value(df_obj: DataFrame, col_name: str, skipna=True) -> float:
        return df_obj.data_frame[col_name].min(skipna=skipna)

    @staticmethod
    def get_max_value(df_obj: DataFrame, col_name: str, skipna=True) -> float:
        """Returns max value in given column"""
        return df_obj.data_frame[col_name].max(skipna=skipna)

    @staticmethod
    def get_mean_value(df_obj: DataFrame, col_name: str, skipna=True) -> float:
        """Returns mean value in given column"""
        return df_obj.data_frame[col_name].mean(skipna=skipna)

    @staticmethod
    def get_trimmed_mean_value(df_obj: DataFrame, col_name: str, lower_bound: float, upper_bound: float) -> float:
        """Returns trimmed mean value in given column"""
        col = df_obj.data_frame[col_name]
        return col[(col >= lower_bound) & (col <= upper_bound)].mean()

    @staticmethod
    def get_variance_value(df_obj: DataFrame, col_name: str, skipna=True) -> float:
        """Returns variance in given column"""
        variance = df_obj.data_frame[col_name].var(skipna=skipna)
        return variance if not pd.isna(variance) else float("nan")

    @staticmethod
    def get_sum_value(df_obj: DataFrame, col_name: str, skipna=True) -> float:
        """Returns sum of values in given column"""
        return df_obj.data_frame[col_name].sum(skipna=skipna)

    @staticmethod
    def get_percentile_values(df_obj: DataFrame, col_name: str, percentiles: List[float]) -> Dict[float, float]:
        """Returns percentile values for given percentiles"""
        ret = {}
        for val in percentiles:
            ret[val] = df_obj.data_frame[col_name].quantile(val / 100.0, interpolation="lower")
        return ret

    @staticmethod
    def get_frequencies(df_obj: DataFrame, col_name: str, values: list = None) -> Dict[Any, int]:
        ret = {}
        if values is not None:
            for val in values:
                if isinstance(val, float) and math.isnan(val):
                    ret[None] = df_obj.data_frame[col_name].isnull().sum()
                else:
                    ret[val] = (df_obj.data_frame[col_name].values == val).sum()
            return ret
        val_df = df_obj.data_frame[col_name].value_counts(dropna=False).to_dict()
        for key, val in val_df.items():
            new_key = key
            if pd.isna(key):
                new_key = None
                if val == 0:
                    continue
            ret[new_key] = val

        return ret

    @staticmethod
    def get_col_df_without_values(df_obj: DataFrame, col_name: str, values: list) -> DataFrame:
        col_df = df_obj.data_frame[col_name]
        col_filter = ~col_df.isin(values)
        nan_excluded = sum(1 for x in values if isinstance(x, float) and math.isnan(x)) > 0
        if nan_excluded:
            col_filter &= ~col_df.isnull()
        new_df = DataFrame()
        new_df.data_frame = col_df[col_filter].to_frame()
        return new_df

    @staticmethod
    def get_size(df_obj: DataFrame):
        return len(df_obj.data_frame)

    @staticmethod
    def get_col_values(df_obj: DataFrame, col_name: str) -> list:
        return df_obj.data_frame[col_name].tolist()

    @staticmethod
    def get_unique_values_in_column(df_obj: DataFrame, col_name: str) -> List:
        """Returns unique values in the given column of dataframe"""
        return df_obj.data_frame[col_name].unique().tolist()
